# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum

POLICY_DAYS_OF_WEEK = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday']


class RandomSampleType(IntEnum):
    DISABLED = 0
    SIZE = 1
    RATIO = 2


class NodeActionType(IntEnum):
    STOP = 0
    START = 1
    WAIT = 2
    EXECUTE = 3


NODE_ACTION_TYPE_NAMES = {
    NodeActionType.STOP: 'stop',
    NodeActionType.START: 'start',
    NodeActionType.WAIT: 'wait',
    NodeActionType.EXECUTE: 'execute'
}


class PodActionType(IntEnum):
    KILL = 0
    WAIT = 1


POD_ACTION_TYPE_NAMES = {
    PodActionType.KILL: 'kill',
    PodActionType.WAIT: 'wait'
}


class PodMatcherTypes(IntEnum):
    NAMESPACE = 0
    DEPLOYMENT = 1
    LABELS = 2


DEFAULT_OUTPUT_NODE_SCENARIO = {
    'name': '',
    'matchers': [],
    'filters': [],
    'actions': [],
    'dayOfWeek': {
        'monday': True,
        'tuesday': True,
        'wednesday': True,
        'thursday': True,
        'friday': True,
        'saturday': False,
        'sunday': False
    },
    'startTime': {
        'hour': 10,
        'minute': 0,
        'second': 0
    },
    'endTime': {
        'hour': 17,
        'minute': 30,
        'second': 0
    },
    'randomSample': {
        'type': RandomSampleType.DISABLED,
        'size': 5,
        'ratio': 0.2
    },
    'probabilityPassAll': {
        'isEnabled': False,
        'probability': 0.5
    },
}

DEFAULT_OUTPUT_POD_SCENARIO = {
    'name': '',
    'matchers': [],
    'filters': [],
    'isTimeOfExecutionEnabled': False,
    'dayOfWeek': {
        'monday': True,
        'tuesday': True,
        'wednesday': True,
        'thursday': True,
        'friday': True,
        'saturday': False,
        'sunday': False
    },
    'startTime': {
        'hour': 10,
        'minute': 0,
        'second': 0
    },
    'endTime': {
        'hour': 17,
        'minute': 30,
        'second': 0
    },
    'randomSample': {
        'type': RandomSampleType.DISABLED,
        'size': 5,
        'ratio': 0.2
    },
    'probabilityPassAll': {
        'isEnabled': False,
        'probability': 0.5
    },
    'actions': []
}




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random

from powerfulseal.metriccollectors.collector import POD_SOURCE
from .scenario import Scenario


class PodScenario(Scenario):
    """ Pod scenario handler.

        Adds metching for k8s-specific things and pod-specific actions
    """

    def __init__(self, name, schema, inventory, k8s_inventory, executor,
                 logger=None, metric_collector=None):
        Scenario.__init__(self, name, schema, logger=logger, metric_collector=metric_collector)
        self.inventory = inventory
        self.k8s_inventory = k8s_inventory
        self.executor = executor

    def match(self):
        """ Makes a union of all the pods matching any of the policy criteria.
        """
        mapping = {
            "namespace": self.match_namespace,
            "deployment": self.match_deployment,
            "labels": self.match_labels,
        }
        selected = set()
        criteria = self.schema.get("match", [])
        for criterion in criteria:
            for key, method in mapping.items():
                if key in criterion:
                    params = criterion.get(key)
                    for pod in method(params):
                        self.logger.debug("Matching %r", pod)
                        selected.add(pod)
        if len(selected) == 0:
            self.metric_collector.add_matched_to_empty_set_metric(POD_SOURCE)
        return list(selected)

    def match_namespace(self, params):
        """ Matches pods for a namespace
        """
        namespace = params.get("name")
        pods = self.k8s_inventory.find_pods(
            namespace=namespace,
        )
        self.logger.info("Matched %d pods in namespace %s", len(pods), namespace)
        return pods

    def match_deployment(self, params):
        """ Matches pods for a deployment
        """
        namespace = params.get("namespace")
        deployment_name = params.get("name")
        pods = self.k8s_inventory.find_pods(
            namespace=namespace,
            deployment_name=deployment_name,
        )
        self.logger.info("Matched %d pods for deploy %s in namespace %s",
                 len(pods), deployment_name, namespace
        )
        return pods

    def match_labels(self, params):
        """ Matches pods for a selector
        """
        namespace = params.get("namespace")
        selector = params.get("selector")
        pods = self.k8s_inventory.find_pods(
            namespace=namespace,
            selector=selector,
        )
        self.logger.info("Matched %d pods for selector %s in namespace %s",
                 len(pods), selector, namespace
        )
        return pods

    def action_kill(self, item, params):
        """ Kills a pod by executing a docker kill on one of the containers or pod delete
        """
        probability = params.get("probability", 1)

        # In case we are setup to delete pods, instead of SSHing
        if self.k8s_inventory.delete_pods:
            if probability >= random.random():
                self.logger.info("Deleting pod %r", item)
                self.k8s_inventory.k8s_client.delete_pods([item])
            else:
                self.logger.info("Pod got lucky - not deleting")
            return

        node = self.inventory.get_node_by_ip(item.host_ip)
        if node is None:
            self.logger.info("Node not found for pod: %s", item)
            return
        force = params.get("force", True)
        signal = "SIGKILL" if force else "SIGTERM"
        container_id = random.choice(item.container_ids)
        cmd = self.executor.get_kill_command(
            signal=signal,
            container_id=container_id.replace("docker://",""),
        )

        if probability >= random.random():
            self.logger.info("Action execute '%s' on %r", cmd, item)
            for value in self.executor.execute(
                cmd, nodes=[node]
            ).values():
                if value["ret_code"] > 0:
                    self.logger.info("Error return code: %s", value)
                    self.metric_collector.add_pod_kill_failed_metric(item)
                else:
                    self.metric_collector.add_pod_killed_metric(item)
        else:
            self.logger.info("Pod got lucky - not killing")

    def act(self, items):
        """ Executes all the supported actions on the list of pods.
        """
        actions = self.schema.get("actions", [])
        mapping = {
            "wait": self.action_wait,
            "kill": self.action_kill,
        }
        return self.act_mapping(items, actions, mapping)




# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from mock import MagicMock

from powerfulseal.policy.node_scenario import NodeScenario
from powerfulseal.policy.pod_scenario import PodScenario
from powerfulseal.policy.scenario import Scenario
from powerfulseal.execute import RemoteExecutor


# Common fixtures
class Dummy():
    pass

def make_dummy_object():
    return Dummy()

@pytest.fixture
def dummy_object():
    return make_dummy_object()


# Scenario fixtures
class NoopScenario(Scenario):

    def match(self):
        return []

    def act(self, items):
        pass


@pytest.fixture
def noop_scenario():
    return NoopScenario(
        name="test scenario",
        schema={}
    )


# Pod Scenario Fixtures
EXAMPLE_POD_SCHEMA = {
    "match": [
        {
            "property": {
                "name": "attr",
                "value": "a.*"
            }
        },
    ]
}


@pytest.fixture
def pod_scenario():
    inventory = MagicMock()
    k8s_inventory = MagicMock()
    k8s_inventory.delete_pods = False
    executor = RemoteExecutor()
    return PodScenario(
        name="test scenario",
        schema=EXAMPLE_POD_SCHEMA,
        inventory=inventory,
        k8s_inventory=k8s_inventory,
        executor=executor,
    )


# Node Scenario Fixtures
EXAMPLE_NODE_SCHEMA = {
    "match": [
        {
            "property": {
                "name": "attr",
                "value": "a.*"
            }
        },
    ]
}


@pytest.fixture
def node_scenario():
    inventory = MagicMock()
    driver = MagicMock()
    executor = MagicMock()
    return NodeScenario(
        name="test scenario",
        schema=EXAMPLE_NODE_SCHEMA,
        inventory=inventory,
        driver=driver,
        executor=executor,
    )



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import random
import threading
import time
from threading import Lock

import jsonschema
import yaml
from flask import Flask, jsonify, request, send_file, render_template
from flask_cors import CORS
from flask_swagger_ui import get_swaggerui_blueprint

from powerfulseal.policy import PolicyRunner
from powerfulseal.policy.node_scenario import NodeScenario
from powerfulseal.policy.pod_scenario import PodScenario
from powerfulseal.web.formatter import PolicyFormatter

# Flask instance and routes
app = Flask(__name__, static_url_path="/static", static_folder="dist/static", template_folder="dist")

# Set up CORS to allow requests originating outside the server for the API
CORS(app, resources={r"/api/*": {"origins": "*"}})

# Serve Swagger
SWAGGER_PATH = '/docs'
API_URL = '/spec.yml'
swaggerui_blueprint = get_swaggerui_blueprint(
    SWAGGER_PATH,
    API_URL,
    config={
        'app_name': 'PowerfulSeal Docs'
    }
)
app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_PATH)

# Singleton instance of the server
server_state = None


@app.route('/spec.yml')
def get_spec():
    return send_file('spec.yml')


@app.route('/api/policy', methods=['GET', 'POST', 'PUT'])
def policy_actions():
    if request.method == 'GET':
        # GET request: returns a JSON representation of the policy file
        policy = server_state.get_policy()
        return jsonify(PolicyFormatter.output_policy(policy))
    elif request.method == 'POST':
        # POST request: returns a YAML representation of the given policy
        input_policy = request.get_json().get('policy', None)
        if input_policy is None:
            return jsonify({'error': 'Policy field missing'}), 400

        try:
            modified_policy = PolicyFormatter.parse_policy(input_policy)
        except KeyError:
            return jsonify({'error': 'Policy not valid'}), 400
        if not PolicyRunner.is_policy_valid(modified_policy):
            return jsonify({'error': 'Policy not valid'}), 400

        return jsonify({
            'policy': yaml.dump(modified_policy, default_flow_style=False)
        })
    elif request.method == 'PUT':
        # PUT request: modify a policy
        input_policy = request.get_json().get('policy', None)
        if input_policy is None:
            return jsonify({'error': 'Policy field missing'}), 400

        try:
            modified_policy = PolicyFormatter.parse_policy(input_policy)
        except KeyError:
            return jsonify({'error': 'Policy not valid'}), 400
        if not PolicyRunner.is_policy_valid(modified_policy):
            return jsonify({'error': 'Policy not valid'}), 400

        try:
            server_state.update_policy(modified_policy)
        except IOError:
            return jsonify({'error': 'Unable to overwrite policy file'}), 500

        return jsonify({}), 200

    return jsonify({}), 501


@app.route('/api/autonomous-mode', methods=['GET', 'POST'])
def autonomousMode():
    """
    GET: Gets the state of autonomous mode
    POST: Sets the state of autonomous mode (state is either start or stop)
    """
    if request.method == 'GET':
        return jsonify({'isStarted': server_state.is_policy_runner_running()})
    elif request.method == 'POST':
        params = request.get_json()
        action = params.get('action', None)
        if action is None:
            return jsonify({'error': 'Action field missing'}), 400

        if action not in ['start', 'stop']:
            return jsonify({'error': 'Action field must either be \'start\' or \'stop\''}), 400

        try:
            if action == 'start':
                if server_state.is_policy_runner_running():
                    return jsonify({'error': 'Policy runner already running'}), 412
                server_state.start_policy_runner()
            else:
                if not server_state.is_policy_runner_running():
                    return jsonify({'error': 'Policy runner already stopped'}), 412
                server_state.stop_policy_runner()
        except RuntimeError:
            return jsonify({'error': 'Policy runner is in an inconsistent state'}), 500

        return jsonify({})

    return jsonify({}), 501


@app.route('/api/logs')
def logs():
    """
    Retrieves the application logs
    """
    # Logs can be retrieves from the offset inclusive (e.g., if the offset is 3,
    # then return from index three (or fourth item in the list) inclusive [3:])
    offset = request.args.get('offset', default=0, type=int)

    if offset < 0:
        return jsonify({'error': 'Offset cannot be negative'}), 400

    with server_state.lock:
        if len(server_state.logs) < offset + 1:
            return jsonify({'logs': []})
        return jsonify({'logs': server_state.logs[offset:]})


@app.route('/api/nodes', methods=['GET', 'POST'])
def update_nodes():
    """
    GET: Gets a list of all nodes
    POST: Starts or stops a node identified by its IP address
    """
    if request.method == 'GET':
        nodes = [{
            'id': node.id,
            'name': node.name,
            'ip': node.ip,
            'az': node.az,
            'groups': node.groups,
            'no': node.no,
            'state': node.state.value
        } for node in server_state.get_nodes()]

        return jsonify({'nodes': nodes})
    elif request.method == 'POST':
        params = request.get_json()
        action = params.get('action', None)
        ip = params.get('ip', None)
        if action is None or ip is None:
            return jsonify({'error': 'Action/IP address fields missing'}), 400

        if action not in ['start', 'stop']:
            return jsonify({'error': 'Invalid action'}), 400

        for node in server_state.get_nodes():
            if node.ip == ip:
                if action == 'start':
                    is_action_successful = server_state.start_node(node)
                else:
                    is_action_successful = server_state.stop_node(node)

                if is_action_successful:
                    return jsonify({}), 200
                else:
                    return jsonify({'error': 'Action on node failed'}, 500)

        return jsonify({'error': 'Node IP address not found'}), 404

    return jsonify({}), 501


@app.route('/api/pods', methods=['GET', 'POST'])
def update_pods():
    """
    GET: Gets a list of all pods
    POST: Kills or force kills a pod by its UID field. Force kills if `is_forced`
    parameter is the string `true`.
    """
    if request.method == 'GET':
        # Pods can be filtered based on their namespace
        # An empty namespace retrieves all pods
        namespace = request.args.get('namespace', default="", type=str)

        pods = [{
            'name': pod.name,
            'namespace': pod.namespace,
            'num': pod.num,
            'uid': pod.uid,
            'host_ip': pod.host_ip,
            'ip': pod.ip,
            'container_ids': pod.container_ids,
            'restart_count': pod.restart_count,
            'state': pod.state,
            'labels': pod.labels
        } for pod in server_state.get_pods() if len(namespace) == 0 or pod.namespace == namespace]

        return jsonify({'pods': pods})
    elif request.method == 'POST':
        params = request.get_json()
        is_forced = params.get('isForced', False) in [True, 'true', '1', 1]
        uid = params.get('uid', None)
        if uid is None:
            return jsonify({'error': 'uid field missing'}), 400

        for pod in server_state.get_pods():
            if pod.uid == uid:
                is_action_successful = server_state.kill_pod(pod, is_forced)
                if is_action_successful:
                    return jsonify({}), 200
                else:
                    return jsonify({'error': 'Action on pod failed'}, 500)

        return jsonify({'error': 'Pod UID not found'}), 404

    return jsonify({}), 501


@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
def catch_all(path):
    return render_template('index.html', baseUrl='%sapi' % request.base_url)


def start_server(host, port):
    app.run(host=host, port=port)


class ThreadedPolicyRunner(threading.Thread):
    def __init__(self, policy, inventory, k8s_inventory, driver, executor,
            stop_event, logger=None, metric_collector=None):
        threading.Thread.__init__(self)
        config = policy.get("config", {})
        self.wait_min = config.get("minSecondsBetweenRuns", 0)
        self.wait_max = config.get("maxSecondsBetweenRuns", 300)

        self.node_scenarios = [
            NodeScenario(
                name=item.get("name"),
                schema=item,
                inventory=inventory,
                driver=driver,
                executor=executor,
                metric_collector=metric_collector,
            )
            for item in policy.get("nodeScenarios", [])
        ]

        self.pod_scenarios = [
            PodScenario(
                name=item.get("name"),
                schema=item,
                inventory=inventory,
                k8s_inventory=k8s_inventory,
                executor=executor,
                metric_collector=metric_collector,
            )
            for item in policy.get("podScenarios", [])
        ]

        self.inventory = inventory
        self.k8s_inventory = k8s_inventory
        self.driver = driver
        self.executor = executor
        self.logger = logger or logging.getLogger(__name__)

        self.stop_event = stop_event
        self.stop_event.clear()

    def run(self):
        while not self.stop_event.is_set():
            try:
                for scenario in self.node_scenarios:
                    scenario.execute()
                for scenario in self.pod_scenarios:
                    scenario.execute()
            except Exception as e:
                logging.error(e)
                self.stop()

            # Repeatedly check whether the stop event has been set in order to
            # prevent blocking the main thread while it waits for the thread to
            # join after stopping
            sleep_time = int(random.uniform(self.wait_min, self.wait_max))
            self.logger.debug("Sleeping for %s seconds", sleep_time)
            while sleep_time > 0 and not self.stop_event.is_set():
                time.sleep(1)
                sleep_time -= 1

            self.inventory.sync()

    def stop(self):
        self.stop_event.set()


class ServerState:
    def __init__(self, policy, inventory, k8s_inventory, driver, executor,
                 server_host, server_port, policy_path, logger=None,
                 metric_collector=None):
        # server_state must be accessed in a global context as Flask works within
        # a module level scope. As a result, there is no intuitive way for Flask
        # API endpoints to access the server state without first making the server
        # state a global singleton variable which Flask can directly access.
        global server_state
        server_state = self

        self.policy = policy
        self.inventory = inventory
        self.k8s_inventory = k8s_inventory
        self.driver = driver
        self.executor = executor
        self.server_host = server_host
        self.server_port = server_port
        self.policy_path = policy_path
        self.metric_collector = metric_collector

        self.logger = logger or logging.getLogger(__name__)
        self.logs = []

        # Due to concurrent requests, it is safer to acquire locks on variables
        # where race conditions could occur.
        self.lock = Lock()

        # Contains an instance of ThreadedPolicyRunner which is only set once
        # autonomous mode is started and cleared once it is stopped.
        # The threading.Event has to be in the main thread so the state of it can
        # be read while the policy runner thread blocks (e.g., sleeps).
        self.policy_runner = None
        self.policy_runner_stop_event = threading.Event()

    def is_policy_valid(self):
        """
        Checks whether the specified policy is valid depending on the schema
        file used in the PolicyRunner class
        """
        schema = PolicyRunner.get_schema()
        try:
            jsonschema.validate(self.policy, schema)
        except jsonschema.ValidationError:
            return False
        return True

    def get_policy(self):
        with self.lock:
            return self.policy

    def update_policy(self, policy):
        """
        :except IOError
        """
        with self.lock:
            with open(self.policy_path, 'w') as f:
                f.write(yaml.dump(policy, default_flow_style=False))
            self.policy = policy

    def get_nodes(self):
        return self.inventory.get_all_nodes()

    def get_pods(self):
        return self.k8s_inventory.get_all_pods()

    def start_node(self, node):
        """
        Starts a node. Returns true on success and false on failure.
        """
        try:
            self.driver.start(node)
            return True
        except:
            self.logger.error("Error starting the machine")
            return False

    def stop_node(self, node):
        """
        Stops a node. Returns true on success and false on failure.
        """
        try:
            self.driver.stop(node)
            return True
        except:
            self.logger.error("Error stopping the machine")
            return False

    def kill_pod(self, pod, is_forced):
        """
        Kills a pod. Returns true on success and false on failure.
        """
        # Get node to execute kill command on
        node = self.inventory.get_node_by_ip(pod.host_ip)
        if node is None:
            return False

        # Format kill command
        signal = "SIGKILL" if is_forced else "SIGTERM"
        container_id = random.choice(pod.container_ids)
        cmd = self.executor.get_kill_command(
            signal=signal,
            container_id=container_id.replace("docker://", ""),
        )

        # Send kill command
        for value in self.executor.execute(cmd, nodes=[node]).values():
            if value["ret_code"] > 0:
                return False

        return True

    def is_policy_runner_running(self):
        with self.lock:
            return self.policy_runner is not None and not self.policy_runner_stop_event.is_set()

    def start_policy_runner(self):
        self.logger.info("Starting policy runner")
        with self.lock:
            if self.policy_runner is not None and not self.policy_runner_stop_event.is_set():
                raise RuntimeError('Policy runner is already running')

            self.policy_runner = ThreadedPolicyRunner(self.policy, self.inventory,
                                                      self.k8s_inventory, self.driver,
                                                      self.executor,
                                                      self.policy_runner_stop_event,
                                                      metric_collector=self.metric_collector)
            self.policy_runner.start()

    def stop_policy_runner(self):
        self.logger.info("Stopping policy runner")
        with self.lock:
            if self.policy_runner is None or self.policy_runner_stop_event.is_set():
                raise RuntimeError('Policy runner is already stopped')

            self.policy_runner_stop_event.set()
            self.policy_runner = None


class ServerStateLogHandler(logging.Handler):
    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        with server_state.lock:
            server_state.logs.append({
                'timestamp': record.created,
                'level': record.levelname,
                'message': record.getMessage()
            })



# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc, six


@six.add_metaclass(abc.ABCMeta)
class AbstractDriver():
    """
        Abstract class representing a cloud driver.
        All concrete drivers should implement this.

        NOTE: node.extIp should be an accessible IP.
              It should the same as node.ip if there is
              no separate public IP.
    """

    @abc.abstractmethod
    def sync(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_by_ip(self, ip):
        pass  # pragma: no cover

    @abc.abstractmethod
    def stop(self, node):
        pass  # pragma: no cover

    @abc.abstractmethod
    def start(self, node):
        pass  # pragma: no cover

    @abc.abstractmethod
    def delete(self, node):
        pass  # pragma: no cover




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import IntEnum


class NodeState(IntEnum):
    """
        Enumeration to describe a Node state
    """
    UNKNOWN = 1
    UP = 2
    DOWN = 3


class Node(object):
    """
        Basic class representing a machine in the cluster
    """


    def __init__(self, id, name=None, ip=None, extIp=None, az=None,
            groups=None, no=None, state=None):
        self.id = id
        self.name = name
        self.ip = ip
        self.extIp = extIp
        self.az = az
        self.groups = groups or []
        self.no = no
        if state is None:
            self.state = NodeState.UNKNOWN
        elif type(state) is not NodeState:
            raise ValueError("Please use a NodeState; %s provided" % state)
        else:
            self.state = state

    def __str__(self):
        return (
            "[node no={no} id={id} ip={ip} extIp={extIp} az={az} groups={groups} name={name} "
            "state={state}]"
        ).format(
            no=self.no,
            id=self.id,
            name=self.name,
            ip=self.ip,
            extIp=self.extIp,
            az=self.az,
            groups=self.groups,
            state=str(self.state)
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return other.id == self.id

    def __hash__(self):
        return hash(self.id)




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from datetime import datetime
from .pod import Pod


class K8sInventory():
    """ Kubernetes inventory - deal with namespaces, deployments and pods.
        Also manages cache.
    """

    def __init__(self, k8s_client, delete_pods=False, logger=None):
        self.k8s_client = k8s_client
        self._cache_namespaces = []
        self._cache_last = None
        self.delete_pods = delete_pods
        self.logger = logger or logging.getLogger(__name__)
        self.last_pods = []

    def is_fresh(self, when):
        """ Helper to invalidate the cache.
        """
        delta = datetime.now() - when
        if delta.seconds > 10:
            return False
        return True

    def find_namespaces(self):
        """ Returns all namespaces.
        """
        if self._cache_last is not None and self.is_fresh(self._cache_last):
            self.logger.info("Using cached namespaces")
            return self._cache_namespaces
        self.logger.info("Reading kubernetes namespaces")
        namespaces = [
            item.metadata.name for item in self.k8s_client.list_namespaces()
        ]
        self._cache_namespaces = namespaces
        self._cache_last = datetime.now()
        return namespaces

    def find_deployments(self, namespace=None, labels=None):
        """ Find deployments for a namespace (default to "default" namespace).
        """
        # Check if namespace is None instead of using "or" as Kubernetes treats
        # an empty string as the */all wildcard
        if namespace is None:
            namespace = "default"
        return [
            item.metadata.name
            for item in self.k8s_client.list_deployments(
                namespace=namespace,
                labels=labels,
            )
        ]

    def find_pods(self, namespace, selector=None, deployment_name=None):
        """ Find pods in a namespace, for a deployment or selector.
        """
        # Check if namespace is None instead of using "or" as Kubernetes treats
        # an empty string as the */all wildcard
        if namespace is None:
            namespace = "default"
        pods = self.k8s_client.list_pods(
            namespace=namespace,
            selector=selector,
            deployment_name=deployment_name,
        )
        pod_objects = []
        if pods is not None:
            for i, item in enumerate(pods):
                        container_ids = []
                        restart_count = []
                        if item.status.container_statuses:
                            for status in item.status.container_statuses:
                                container_ids.append(status.container_id)
                                restart_count.append(status.restart_count)
                        
                        pod_objects.append(Pod(
                            num=i,
                            name=item.metadata.name,
                            namespace=item.metadata.namespace,
                            uid=item.metadata.uid,
                            host_ip=item.status.host_ip,
                            ip=item.status.pod_ip,
                            container_ids=container_ids,
                            restart_count=sum(restart_count),
                            state=item.status.phase,
                            labels=item.metadata.labels,
                            meta=item,
                        ))
        self.last_pods = pod_objects
        return pod_objects

    def get_all_pods(self):
        """
        Retrieves all pods for all namespaces
        """
        return self.find_pods("")



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

from powerfulseal.web.constants import *


class PolicyFormatter:
    """
    Converts between internal representation of a policy and the representation
    used by the API
    """
    @classmethod
    def output_policy(cls, policy):
        output_node_scenarios = []
        for policy_scenario in policy.get('nodeScenarios', []):
            output_node_scenarios.append(PolicyFormatter.output_node_scenario(policy_scenario))

        # Format pod scenario output for JSON output
        output_pod_scenarios = []
        for policy_scenario in policy.get('podScenarios', []):
            output_pod_scenarios.append(PolicyFormatter.output_pod_scenario(policy_scenario))

        return {
            'minSecondsBetweenRuns': policy.get('config', {}).get('minSecondsBetweenRuns', 0),
            'maxSecondsBetweenRuns': policy.get('config', {}).get('maxSecondsBetweenRuns', 300),
            'nodeScenarios': output_node_scenarios,
            'podScenarios': output_pod_scenarios
        }

    @classmethod
    def output_node_scenario(cls, policy_scenario):
        output_node_scenario = copy.deepcopy(DEFAULT_OUTPUT_NODE_SCENARIO)
        output_node_scenario['name'] = policy_scenario['name']

        # Process matchers
        for matcherItem in policy_scenario.get('match', []):
            output_node_scenario['matchers'].append(matcherItem['property'])

        # Process filters
        for filterItem in policy_scenario.get('filters', []):
            # In the schema, filter items are represented by an object with a single item

            # Process property filters
            if 'property' in filterItem:
                output_node_scenario['filters'].append(filterItem['property'])

            # Process time of execution filters
            if 'dayTime' in filterItem:
                for dayOfWeek in POLICY_DAYS_OF_WEEK:
                    output_node_scenario['dayOfWeek'][dayOfWeek] = \
                        dayOfWeek in filterItem['dayTime']['onlyDays']
                output_node_scenario['startTime'] = filterItem['dayTime']['startTime']
                output_node_scenario['endTime'] = filterItem['dayTime']['endTime']
                output_node_scenario['isTimeOfExecutionEnabled'] = True

            # Process random sample filters
            if 'randomSample' in filterItem:
                if 'size' in filterItem['randomSample']:
                    # The default scenario's other keys must not be modified or else the front-end
                    # will missed keys which are being watched
                    output_node_scenario['randomSample']['type'] = RandomSampleType.SIZE.value
                    output_node_scenario['randomSample']['size'] = filterItem['randomSample']['size']
                elif 'ratio' in filterItem['randomSample']:
                    output_node_scenario['randomSample']['type'] = RandomSampleType.RATIO.value
                    output_node_scenario['randomSample']['ratio'] = filterItem['randomSample']['ratio']

            # Process probability pass all filters
            if 'probability' in filterItem:
                output_node_scenario['probabilityPassAll']['isEnabled'] = True
                output_node_scenario['probabilityPassAll']['probability'] = \
                    filterItem['probability']['probabilityPassAll']

        # Process actions
        for actionItem in policy_scenario.get('actions', []):
            if 'stop' in actionItem:
                output_node_scenario['actions'].append({
                    'type': NodeActionType.STOP.value,
                    'params': [{
                        'name': k,
                        'value': v,
                    } for (k, v) in actionItem['stop'].items()]
                })
            elif 'start' in actionItem:
                output_node_scenario['actions'].append({
                    'type': NodeActionType.START.value,
                    'params': []
                })
            elif 'wait' in actionItem:
                output_node_scenario['actions'].append({
                    'type': NodeActionType.WAIT.value,
                    'params': [
                        {
                            'name': 'seconds',
                            'value': actionItem['wait']['seconds']
                        }
                    ]
                })
            elif 'execute' in actionItem:
                output_node_scenario['actions'].append({
                    'type': NodeActionType.EXECUTE.value,
                    'params': [
                        {
                            'name': 'cmd',
                            'value': actionItem['execute']['cmd']
                        }
                    ]
                })

        return output_node_scenario

    @classmethod
    def output_pod_scenario(cls, policy_scenario):
        output_pod_scenario = copy.deepcopy(DEFAULT_OUTPUT_POD_SCENARIO)
        output_pod_scenario['name'] = policy_scenario['name']

        # Process matchers
        output_pod_scenario['matchers'] = []
        for matcherItem in policy_scenario.get('match', []):
            if 'namespace' in matcherItem:
                output_pod_scenario['matchers'].append({
                    'type': PodMatcherTypes.NAMESPACE.value,
                    'params': [{
                        'name': k,
                        'value': v
                    } for (k, v) in matcherItem['namespace'].items()]
                })
            elif 'deployment' in matcherItem:
                output_pod_scenario['matchers'].append({
                    'type': PodMatcherTypes.DEPLOYMENT.value,
                    'params': [{
                        'name': k,
                        'value': v
                    } for (k, v) in matcherItem['deployment'].items()]
                })
            elif 'labels' in matcherItem:
                output_pod_scenario['matchers'].append({
                    'type': PodMatcherTypes.LABELS.value,
                    'params': [{
                        'name': k,
                        'value': v,
                    } for (k, v) in matcherItem['labels'].items()]
                })

        # Process filters
        output_pod_scenario['filters'] = []
        for filterItem in policy_scenario.get('filters', []):
            # Process property filters
            if 'property' in filterItem:
                output_pod_scenario['filters'].append(filterItem['property'])

            # Process time of execution filters
            if 'dayTime' in filterItem:
                for dayOfWeek in POLICY_DAYS_OF_WEEK:
                    output_pod_scenario['dayOfWeek'][dayOfWeek] = \
                        dayOfWeek in filterItem['dayTime']['onlyDays']
                output_pod_scenario['startTime'] = filterItem['dayTime']['startTime']
                output_pod_scenario['endTime'] = filterItem['dayTime']['endTime']
                output_pod_scenario['isTimeOfExecutionEnabled'] = True

            # Process random sample filters
            if 'randomSample' in filterItem:
                if 'size' in filterItem['randomSample']:
                    # The default scenario's other keys must not be modified or else the front-end
                    # will missed keys which are being watched
                    output_pod_scenario['randomSample']['type'] = RandomSampleType.SIZE.value
                    output_pod_scenario['randomSample']['size'] = filterItem['randomSample']['size']
                elif 'ratio' in filterItem['randomSample']:
                    output_pod_scenario['randomSample']['type'] = RandomSampleType.RATIO.value
                    output_pod_scenario['randomSample']['ratio'] = filterItem['randomSample']['ratio']

            # Process probability pass all filters
            if 'probability' in filterItem:
                output_pod_scenario['probabilityPassAll']['isEnabled'] = True
                output_pod_scenario['probabilityPassAll']['probability'] = \
                    filterItem['probability']['probabilityPassAll']

        # Process actions
        output_pod_scenario['actions'] = []
        for actionItem in policy_scenario.get('actions', []):
            if 'kill' in actionItem:
                output_pod_scenario['actions'].append({
                    'type': PodActionType.KILL.value,
                    'params': [{
                        'name': k,
                        'value': v,
                    } for (k, v) in actionItem['kill'].items()]
                })
            elif 'wait' in actionItem:
                output_pod_scenario['actions'].append({
                    'type': PodActionType.WAIT.value,
                    'params': [
                        {
                            'name': 'seconds',
                            'value': actionItem['wait']['seconds']
                        }
                    ]
                })

        return output_pod_scenario

    @classmethod
    def parse_policy(cls, input_policy):
        """
        :raises KeyError on parse failure
        """
        return {
            'config': {
                'minSecondsBetweenRuns': input_policy.get('minSecondsBetweenRuns'),
                'maxSecondsBetweenRuns': input_policy.get('maxSecondsBetweenRuns')
            },
            'nodeScenarios': [
                PolicyFormatter.parse_node_scenario(node_scenario)
                for node_scenario in input_policy.get('nodeScenarios', [])
            ],
            'podScenarios': [
                PolicyFormatter.parse_pod_scenario(pod_scenario)
                for pod_scenario in input_policy.get('podScenarios', [])
            ]
        }

    @classmethod
    def parse_node_scenario(cls, input_scenario):
        """
        :raises KeyError on parse failure
        """
        policy_scenario = {
            'name': input_scenario['name']
        }

        # Process matchers
        policy_scenario['match'] = []
        for matcher in input_scenario.get('matchers', []):
            policy_scenario['match'].append({
                'property': matcher
            })

        # Process filter properties
        policy_scenario['filters'] = []
        for filter in input_scenario.get('filters', []):
            policy_scenario['filters'].append({
                'property': filter
            })

        # Process time of execution filters
        if input_scenario.get('isTimeOfExecutionEnabled'):
            policy_scenario['filters'].append({
                'dayTime': {
                    'onlyDays': [dayOfWeek for dayOfWeek in POLICY_DAYS_OF_WEEK if
                                 input_scenario['dayOfWeek'][dayOfWeek]],
                    'startTime': input_scenario['startTime'],
                    'endTime': input_scenario['endTime']
                }
            })

        # Process random sample
        if input_scenario['randomSample']['type'] == RandomSampleType.SIZE.value:
            policy_scenario['filters'].append({
                'randomSample': {
                    'size': input_scenario['randomSample']['size']
                }
            })
        elif input_scenario['randomSample']['type'] == RandomSampleType.RATIO.value:
            policy_scenario['filters'].append({
                'randomSample': {
                    'ratio': input_scenario['randomSample']['ratio']
                }
            })

        # Process probability pass all
        if input_scenario['probabilityPassAll']['isEnabled']:
            policy_scenario['filters'].append({
                'probability': {
                    'probabilityPassAll': input_scenario['probabilityPassAll']['probability']
                }
            })

        # Process actions
        policy_scenario['actions'] = []
        for action in input_scenario.get('actions', []):
            policy_scenario['actions'].append({
                NODE_ACTION_TYPE_NAMES[action['type']]: {param['name']: param['value'] for param in action['params']}
            })

        return policy_scenario

    @classmethod
    def parse_pod_scenario(cls, input_scenario):
        policy_scenario = {
            'name': input_scenario['name']
        }

        # Process matchers
        policy_scenario['match'] = []
        for matcher in input_scenario.get('matchers', []):
            if matcher['type'] == PodMatcherTypes.NAMESPACE.value:
                policy_scenario['match'].append({
                    'namespace': {param['name']: param['value'] for param in matcher['params']}
                })
            elif matcher['type'] == PodMatcherTypes.DEPLOYMENT.value:
                policy_scenario['match'].append({
                    'deployment': {param['name']: param['value'] for param in matcher['params']}
                })
            elif matcher['type'] == PodMatcherTypes.LABELS.value:
                policy_scenario['match'].append({
                    'labels': {param['name']: param['value'] for param in matcher['params']}
                })

        # Process filter properties
        policy_scenario['filters'] = []
        for filter_item in input_scenario.get('filters', []):
            policy_scenario['filters'].append({
                'property': filter_item
            })

        # Process time of execution filters
        if input_scenario['isTimeOfExecutionEnabled']:
            policy_scenario['filters'].append({
                'dayTime': {
                    'onlyDays': [dayOfWeek for dayOfWeek in POLICY_DAYS_OF_WEEK if
                                 input_scenario['dayOfWeek'][dayOfWeek]],
                    'startTime': input_scenario['startTime'],
                    'endTime': input_scenario['endTime']
                }
            })

        # Process random sample
        if input_scenario['randomSample']['type'] == RandomSampleType.SIZE.value:
            policy_scenario['filters'].append({
                'randomSample': {
                    'size': input_scenario['randomSample']['size']
                }
            })
        elif input_scenario['randomSample']['type'] == RandomSampleType.RATIO.value:
            policy_scenario['filters'].append({
                'randomSample': {
                    'ratio': input_scenario['randomSample']['ratio']
                }
            })

        # Process probability pass all
        if input_scenario['probabilityPassAll']['isEnabled']:
            policy_scenario['filters'].append({
                'probability': {
                    'probabilityPassAll': input_scenario['probabilityPassAll']['probability']

                }
            })

        # Process actions
        policy_scenario['actions'] = []
        for action in input_scenario['actions']:
            policy_scenario['actions'].append({
                POD_ACTION_TYPE_NAMES[action['type']]: {param['name']: param['value'] for param in action['params']}
            })

        return policy_scenario



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random

import pytest
from mock import MagicMock
from prometheus_client import REGISTRY

from powerfulseal.metriccollectors import PrometheusCollector
from powerfulseal.metriccollectors.collector import NODE_SOURCE, POD_SOURCE
from powerfulseal.metriccollectors.prometheus_collector import STATUS_SUCCESS, \
    STATUS_FAILURE, POD_KILLS_METRIC_NAME, NODE_STOPS_METRIC_NAME, \
    EXECUTE_FAILED_METRIC_NAME, FILTERED_TO_EMPTY_SET_METRIC_NAME, \
    PROBABILITY_FILTER_NOT_PASSED_METRIC_NAME, MATCHED_TO_EMPTY_SET_METRIC_NAME
# noinspection PyUnresolvedReferences
from tests.fixtures import node_scenario, make_dummy_object
# noinspection PyUnresolvedReferences
from tests.fixtures import noop_scenario
# noinspection PyUnresolvedReferences
from tests.fixtures import pod_scenario


@pytest.fixture
def prometheus_noop_scenario(noop_scenario):
    noop_scenario.metric_collector = PrometheusCollector()
    return noop_scenario


@pytest.fixture
def prometheus_pod_scenario(pod_scenario):
    pod_scenario.metric_collector = PrometheusCollector()
    return pod_scenario


@pytest.fixture
def prometheus_node_scenario(node_scenario):
    node_scenario.metric_collector = PrometheusCollector()
    return node_scenario


def get_aggregated_sample_values(registry, name, labels=None):
    """
    Returns the aggregated value based on cases where all the labels specified
    match. Returns 0 if no values are found.

    The reason why REGISTRY.get_sample_value is not directly used is that the
    library function performs an equality on all the labels instead of only the
    labels provided.
    """
    labels = labels or {}
    total = 0
    for metric in registry.collect():
        for n, l, value in metric.samples:
            if n == name:
                # Ensure that all labels provided match
                matching = all(k in l and v == l[k] for _, (k, v) in enumerate(labels.items()))
                if matching:
                    total += value
    return total


def test_add_pod_killed_metric(prometheus_pod_scenario):
    prometheus_pod_scenario.schema = {
        "actions": [
            {
                "kill": {
                    "force": False
                }
            },
        ]
    }
    magic_mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 0
        },
    })

    prometheus_pod_scenario.executor.execute = magic_mock

    mock_item1 = MagicMock()
    mock_item1.container_ids = ["docker://container1"]
    mock_item1.namespace = "default"
    mock_item1.name = "testpod1"

    mock_item2 = MagicMock()
    mock_item2.container_ids = ["docker://container2"]
    mock_item2.namespace = "default"
    mock_item2.name = "testpod2"

    before = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_SUCCESS})
    assert before == 0

    prometheus_pod_scenario.act([mock_item1, mock_item1, mock_item1, mock_item2])

    after_total = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_SUCCESS})
    assert after_total == 4

    after_first = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_SUCCESS,
                                                                                        'namespace': 'default',
                                                                                        'name': 'testpod1'})
    assert after_first == 3

    after_second = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_SUCCESS,
                                                                                         'namespace': 'default',
                                                                                         'name': 'testpod2'})
    assert after_second == 1


def test_add_pod_kill_failed_metric(prometheus_pod_scenario):
    prometheus_pod_scenario.schema = {
        "actions": [
            {
                "kill": {
                    "force": False
                }
            },
        ]
    }
    magic_mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 1
        },
    })
    prometheus_pod_scenario.executor.execute = magic_mock

    mock_item1 = MagicMock()
    mock_item1.container_ids = ["docker://container1"]
    mock_item1.namespace = "default"
    mock_item1.name = "testpod1"

    mock_item2 = MagicMock()
    mock_item2.container_ids = ["docker://container2"]
    mock_item2.namespace = "default"
    mock_item2.name = "testpod2"

    before = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_FAILURE})
    assert before == 0

    prometheus_pod_scenario.act([mock_item1, mock_item1, mock_item1, mock_item2])

    after_total = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_FAILURE})
    assert after_total == 4

    after_first = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_FAILURE,
                                                                                        'namespace': 'default',
                                                                                        'name': 'testpod1'})
    assert after_first == 3

    after_second = get_aggregated_sample_values(REGISTRY, POD_KILLS_METRIC_NAME, labels={'status': STATUS_FAILURE,
                                                                                         'namespace': 'default',
                                                                                         'name': 'testpod2'})
    assert after_second == 1


def test_add_node_stopped_metric(prometheus_node_scenario):
    prometheus_node_scenario.schema = {
        "actions": [
            {
                "stop": {}
            },
        ],
    }

    mock_item1 = MagicMock()
    mock_item1.id = "1"
    mock_item1.name = "node1"

    mock_item2 = MagicMock()
    mock_item2.id = "2"
    mock_item2.name = "node2"

    items = [mock_item1, mock_item2]

    before = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_SUCCESS})
    assert before == 0

    prometheus_node_scenario.act(items)

    after_total = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_SUCCESS})
    assert after_total == 2

    after_first = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_SUCCESS,
                                                                                         'id': '1',
                                                                                         'name': 'node1'})
    assert after_first == 1

    after_second = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_SUCCESS,
                                                                                          'id': '2',
                                                                                          'name': 'node2'})
    assert after_second == 1


def test_add_node_stop_failed_metric(prometheus_node_scenario):
    prometheus_node_scenario.schema = {
        "actions": [
            {
                "stop": {}
            },
        ],
    }
    method = getattr(prometheus_node_scenario.driver, "stop")
    method.side_effect = Exception("something bad")

    mock_item1 = MagicMock()
    mock_item1.id = "1"
    mock_item1.name = "node1"

    mock_item2 = MagicMock()
    mock_item2.id = "2"
    mock_item2.name = "node2"

    items = [mock_item1, mock_item2]

    before = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_FAILURE})
    assert before == 0

    prometheus_node_scenario.act(items)

    after_total = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_FAILURE})
    assert after_total == 2

    after_first = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_FAILURE,
                                                                                         'id': '1',
                                                                                         'name': 'node1'})
    assert after_first == 1

    after_second = get_aggregated_sample_values(REGISTRY, NODE_STOPS_METRIC_NAME, labels={'status': STATUS_FAILURE,
                                                                                          'id': '2',
                                                                                          'name': 'node2'})
    assert after_second == 1


def test_add_execute_failed_metric(prometheus_node_scenario):
    prometheus_node_scenario.schema = {
        "actions": [
            {
                "execute": {
                    "cmd": "test",
                }
            },
        ]
    }
    magic_mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 1
        },
    })

    prometheus_node_scenario.executor.execute = magic_mock

    mock_item = MagicMock()
    mock_item.id = '1'
    mock_item.name = 'node1'
    items = [mock_item]

    before = get_aggregated_sample_values(REGISTRY, EXECUTE_FAILED_METRIC_NAME)
    assert before == 0

    prometheus_node_scenario.act(items)

    after = get_aggregated_sample_values(REGISTRY, EXECUTE_FAILED_METRIC_NAME, labels={'id': '1', 'name': 'node1'})
    assert after == 1


def test_add_filtered_to_empty_set_metric(prometheus_noop_scenario):
    before = REGISTRY.get_sample_value(FILTERED_TO_EMPTY_SET_METRIC_NAME)
    assert before == 0

    prometheus_noop_scenario.execute()

    after = REGISTRY.get_sample_value(FILTERED_TO_EMPTY_SET_METRIC_NAME)
    assert after == 1


def test_add_probability_filter_passed_no_nodes_metric(prometheus_noop_scenario):
    """
    Ensures that add_probability_filter_passed_no_nodes_metric is called when
    the filter decides to pass no nodes based on a probability
    """
    assert prometheus_noop_scenario.name == "test scenario"
    random.seed(6)  # make the tests deterministic
    candidates = [make_dummy_object()]

    before = REGISTRY.get_sample_value(PROBABILITY_FILTER_NOT_PASSED_METRIC_NAME)
    assert before == 0

    criterion = {"probabilityPassAll": 0.00000001}
    prometheus_noop_scenario.filter_probability(candidates, criterion)

    after = REGISTRY.get_sample_value(PROBABILITY_FILTER_NOT_PASSED_METRIC_NAME)
    assert after == 1


def test_add_matched_to_empty_set_metric(prometheus_node_scenario, prometheus_pod_scenario):
    prometheus_node_scenario.schema = {
        "match": [
            {
                "namespace": {
                    "name": "non-existent"
                }
            }
        ]
    }

    before = get_aggregated_sample_values(REGISTRY, MATCHED_TO_EMPTY_SET_METRIC_NAME)
    assert before == 0

    prometheus_node_scenario.match()

    after = get_aggregated_sample_values(REGISTRY, MATCHED_TO_EMPTY_SET_METRIC_NAME, labels={'source': NODE_SOURCE})
    assert after == 1

    prometheus_pod_scenario.schema = {
        "match": [
            {
                "namespace": {
                    "name": "non-existent"
                }
            }
        ]
    }

    before = get_aggregated_sample_values(REGISTRY, MATCHED_TO_EMPTY_SET_METRIC_NAME)
    assert before == 1

    prometheus_pod_scenario.match()

    after = get_aggregated_sample_values(REGISTRY, MATCHED_TO_EMPTY_SET_METRIC_NAME, labels={'source': POD_SOURCE})
    assert after == 1

    after_total = get_aggregated_sample_values(REGISTRY, MATCHED_TO_EMPTY_SET_METRIC_NAME)
    assert after_total == 2




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from powerfulseal.metriccollectors.collector import NODE_SOURCE
from .scenario import Scenario


class NodeScenario(Scenario):
    """ NodeScenarios scenario handler.

        Adds metching for nodes and node-specific actions
    """

    def __init__(self, name, schema, inventory, driver,
                 executor, logger=None, metric_collector=None):
        Scenario.__init__(self, name, schema, logger=logger, metric_collector=metric_collector)
        self.inventory = inventory
        self.driver = driver
        self.executor = executor

    def match(self):
        """ Makes a union of all the nodes matching any of the policy criteria.
        """
        selected_nodes = set()
        criteria = self.schema.get("match", [])
        for node in self.inventory.find_nodes():
            for criterion in criteria:
                match = criterion.get("property")
                if self.match_property(node, match):
                    self.logger.debug("Matching %r", node)
                    selected_nodes.add(node)
        if len(selected_nodes) == 0:
            self.metric_collector.add_matched_to_empty_set_metric(NODE_SOURCE)
        return list(selected_nodes)

    def action_start(self, item, params):
        """ Action to start a node.
        """
        self.logger.info("Action start on %r", item)
        try:
            self.driver.start(item)
        except:
            self.logger.exception("Error starting the machine")

    def action_stop(self, item, params):
        """ Action to stop a node.
        """
        self.logger.info("Action stop on %r", item)
        try:
            self.metric_collector.add_node_stopped_metric(item)
            self.driver.stop(item)
        except:
            self.metric_collector.add_node_stop_failed_metric(item)
            self.logger.exception("Error stopping the machine")

    def action_execute(self, item, params):
        """ Executes arbitrary code on the node.
        """
        cmd = params.get("cmd", "hostname")
        self.logger.info("Action execute '%s' on %r", cmd, item)
        for value in self.executor.execute(
            cmd, nodes=[item]
        ).values():
            if value["ret_code"] > 0:
                self.logger.info("Error return code: %s", value)
                self.metric_collector.add_execute_failed_metric(item)

    def act(self, items):
        """ Executes all the supported actions on the list of nodes.
        """
        self.logger.info("Acting on these: %r", items)
        actions = self.schema.get("actions", [])
        mapping = {
            "stop": self.action_stop,
            "start": self.action_start,
            "wait": self.action_wait,
            "execute": self.action_execute,
        }
        return self.act_mapping(items, actions, mapping)




import logging
import boto3
from . import AbstractDriver
from ..node import Node, NodeState


def create_connection_from_config():
    """ Creates a new aws api connection """
    conn = boto3.resource('ec2')
    return conn

def get_all_ips(instance):
    """ Returns the public and private ip addresses of an AWS EC2 instances
    """
    output = []
    output.append(instance.private_ip_address)
    output.append(instance.public_ip_address)
    return output

#https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_InstanceState.html
MAPPING_STATES_STATUS = {
    "RUNNING": NodeState.UP,
    "STOPPED": NodeState.DOWN,
    "TERMINATED": NodeState.DOWN,
}

def server_status_to_state(status):
    return MAPPING_STATES_STATUS.get(status['Name'].upper(), NodeState.UNKNOWN)

def create_node_from_server(server):
    """ Translate AWS EC2 Instance representation into a Node object.
    """
    return Node(
        id=server.id,
        ip=server.private_ip_address,
        extIp=server.public_ip_address,
        az=server.placement['AvailabilityZone'],
        name="",
        state=server_status_to_state(server.state),
)

class AWSDriver(AbstractDriver):
    """
        Concrete implementation of the AWS cloud driver.
    """

    def __init__(self, cloud=None, conn=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.conn = create_connection_from_config()
        self.instances = []

    def sync(self):
        """ Downloads a fresh set of nodes form the API.
        """
        self.logger.info("Synchronizing remote nodes")
        self.remote_servers = self.conn.instances.all()
        self.amount_of_servers = list(self.conn.instances.all())
        self.logger.info("Fetched %s remote servers" % len(self.amount_of_servers))

    def get_by_ip(self, ip):
        """ Retrieve an instance of Node by its IP.
        """
        for server in self.remote_servers:
            addresses = get_all_ips(server)
            if not addresses:
                self.logger.warning("No ip addresses found: %s", server)
            else:
                for addr in addresses:
                    if addr == ip:
                        return create_node_from_server(server)
        return None

    def stop(self, node):
        """ Stop a Node.
        """
        self.conn.instances.filter(InstanceIds=(node.id.split())).stop()

    def start(self, node):
        """ Start a Node.
        """
        self.conn.instances.filter(InstanceIds=(node.id.split())).start()

    def delete(self, node):
        """ Delete a Node permanently.
        """
        self.conn.instances.filter(InstanceIds=(node.id.split())).terminate()




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import logging

import spur


class RemoteExecutor(object):
    """ Executes commands on Node instances via SSH.
        Assumes password-less setup.
    """

    PREFIX = ["sh", "-c"]
    DEFAULT_KILL_COMMAND = "sudo docker kill -s {signal} {container_id}"

    def __init__(self, nodes=None, user="cloud-user",
                 ssh_allow_missing_host_keys=False, ssh_path_to_private_key=None,
                 ssh_password=None, ssh_kill_command=None, override_host=None,
                 use_private_ip=False, logger=None):
        self.nodes = nodes or []
        self.user = user
        self.missing_host_key = (spur.ssh.MissingHostKey.accept
                                 if ssh_allow_missing_host_keys
                                 else spur.ssh.MissingHostKey.raise_error)
        self.ssh_path_to_private_key = ssh_path_to_private_key
        self.ssh_password = ssh_password
        self.ssh_kill_command = ssh_kill_command or self.DEFAULT_KILL_COMMAND
        self.override_host = override_host
        self.use_private_ip = use_private_ip
        self.logger = logger or logging.getLogger(__name__)

    def execute(self, cmd, nodes=None, use_private_ip=None, debug=False):
        """
            Executes an arbitrary command, prefixed with sh -c, on each node
        """
        nodes = nodes or self.nodes
        use_private_ip = use_private_ip or self.use_private_ip
        results = dict()
        cmd_full = self.PREFIX + [cmd]
        for node in nodes:
            hostip = node.extIp
            if use_private_ip:
                hostip = node.ip
            if self.override_host:
                hostip = self.override_host
            if self.ssh_password is not None:
                shell = spur.SshShell(
                    hostname=hostip,
                    username=self.user,
                    missing_host_key=self.missing_host_key,
                    password=self.ssh_password,
                )
            else:
                shell = spur.SshShell(
                    hostname=hostip,
                    username=self.user,
                    missing_host_key=self.missing_host_key,
                    private_key_file=self.ssh_path_to_private_key,
                )

            self.logger.info("Executing '%s' on %s" % (cmd_full, node.name))
            try:
                with shell:
                    output = shell.run(cmd_full)
                    results[hostip] = {
                        "ret_code": output.return_code,
                        "stdout": output.output.decode('utf-8'),
                        "stderr": output.stderr_output.decode('utf-8'),
                    }
            except Exception as e:
                results[hostip] = {
                    "ret_code": 1,
                    "error": str(e),
                }
                self.logger.info("Executing '%s' on %s failed with error: %s" % (cmd_full, node.name, str(e)))
        return results

    def get_kill_command(self, container_id, signal="SIGKILL"):
        """ Produces a templated command to execute """
        return self.ssh_kill_command.format(signal=str(signal), container_id=str(container_id))



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from threading import Lock

import pkg_resources
import pytest
from mock import mock, MagicMock

from powerfulseal.execute import RemoteExecutor
from powerfulseal.k8s import Pod
from powerfulseal.node import Node, NodeState
from powerfulseal.policy import PolicyRunner
from powerfulseal.web import server
from powerfulseal.web.formatter import PolicyFormatter
from powerfulseal.web.server import ServerState


@pytest.fixture
def client():
    client = server.app.test_client()
    yield client


def test_node_state_values_are_expected():
    """
    The API depends on the integer values of the NodeState enum. If these values
    are changed, then the API spec and front-end need to reflect these changes.
    """
    assert NodeState.UNKNOWN == 1
    assert NodeState.UP == 2
    assert NodeState.DOWN == 3
    assert len(NodeState) == 3


def test_autonomous_mode_integration(client):
    policy = {
        'config': {
            'minSecondsBetweenRuns': 0,
            'maxSecondsBetweenRuns': 2
        },
        'nodeScenarios': [
            {
                'name': 'Node Test'
            }
        ],
        'podScenarios': [
            {
                'name': 'Pod Test'
            }
        ]
    }
    test_inventory = MagicMock()
    test_inventory.sync = MagicMock(return_value=None)

    server_state = ServerState(policy, test_inventory, None, None, RemoteExecutor(), None, None, None)

    # Autonomous mode has not yet started
    result = client.get('/api/autonomous-mode')
    assert json.loads(result.data.decode("utf-8"))['isStarted'] is False

    # Autonomous mode has not yet started so it cannot be stopped
    result = client.post('/api/autonomous-mode', data=json.dumps({
        'action': 'stop'
    }), content_type='application/json')
    assert result.status_code == 412

    # Start autonomous mode
    result = client.post('/api/autonomous-mode', data=json.dumps({
        'action': 'start'
    }), content_type='application/json')
    assert result.status_code == 200

    # Autonomous mode has started
    result = client.get('/api/autonomous-mode')
    assert json.loads(result.data.decode("utf-8"))['isStarted'] is True

    # Autonomous mode has started so it cannot be started
    result = client.post('/api/autonomous-mode', data=json.dumps({
        'action': 'start'
    }), content_type='application/json')
    assert result.status_code == 412

    # Stop autonomous mode
    result = client.post('/api/autonomous-mode', data=json.dumps({
        'action': 'stop'
    }), content_type='application/json')
    assert result.status_code == 200

    # Autonomous mode has stopped
    result = client.get('/api/autonomous-mode')
    assert json.loads(result.data.decode("utf-8"))['isStarted'] is False


def test_get_policy_actions(client):
    server_state_mock = MagicMock()
    server_state_mock.get_policy = MagicMock(return_value={})
    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        result = client.get("/api/policy")
        assert json.loads(result.data.decode("utf-8")) == {
            'minSecondsBetweenRuns': 0,
            'maxSecondsBetweenRuns': 300,
            'nodeScenarios': [],
            'podScenarios': []
        }


def test_post_policy_actions(client):
    server_state_mock = MagicMock()

    valid_policy = {
        'minSecondsBetweenRuns': 0,
        'maxSecondsBetweenRuns': 300,
        'nodeScenarios': [],
        'podScenarios': []
    }

    invalid_policy = {
        'config': {
            'minSecondsBetweenRuns': 'invalid'
        }
    }

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        result = client.post("/api/policy", data=json.dumps({
            'policy': PolicyFormatter.output_policy(valid_policy)
        }), content_type='application/json')
        assert result.status_code == 200

        result = client.post("/api/policy", data=json.dumps({
            'policy': PolicyFormatter.output_policy(invalid_policy)
        }), content_type='application/json')
        assert result.status_code == 400


def test_put_policy_actions(client):
    server_state_mock = MagicMock()
    server_state_mock.update_policy = MagicMock()

    valid_policy_path = pkg_resources.resource_filename("tests.policy", "example_config.yml")
    valid_policy = PolicyRunner.load_file(valid_policy_path)

    invalid_policy = {
        'config': {
            'minSecondsBetweenRuns': 'invalid'
        }
    }

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        result = client.put("/api/policy", data=json.dumps({
            'policy': PolicyFormatter.output_policy(valid_policy)
        }), content_type='application/json')
        assert result.status_code == 200

        result = client.put("/api/policy", data=json.dumps({
            'policy': invalid_policy
        }), content_type='application/json')
        assert result.status_code == 400


def test_get_logs(client):
    # An instance of Lock will have to be created as there is no straightforward
    # way to mock a lock
    server_state_mock = MagicMock()
    server_state_mock.lock = Lock()

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        # Test case where an offset is not specified
        server_state_mock.logs = [str(i) for i in range(3)]
        result = client.get("/api/logs")
        assert json.loads(result.data.decode("utf-8"))['logs'] == [str(i) for i in range(3)]

        # Test case where offset is specified and correct result is given
        server_state_mock.logs = [str(i) for i in range(9)]
        result = client.get("/api/logs?offset=5")
        assert json.loads(result.data.decode("utf-8"))['logs'] == ['5', '6', '7', '8']

        # Test edge case where just within range
        server_state_mock.logs = [str(i) for i in range(9)]
        result = client.get("/api/logs?offset=8")
        assert json.loads(result.data.decode("utf-8"))['logs'] == ['8']

        # Test edge case where just outside range
        server_state_mock.logs = [str(i) for i in range(9)]
        result = client.get("/api/logs?offset=9")
        assert json.loads(result.data.decode("utf-8"))['logs'] == []

        # Test case where offset is negative
        # Test edge case where just within range
        server_state_mock.logs = [str(i) for i in range(9)]
        result = client.get("/api/logs?offset=-1")
        assert result.status_code == 400


def test_get_nodes(client):
    server_state_mock = MagicMock()

    test_node = Node(1, name='a',
                     ip='0.0.0.0',
                     az='A',
                     groups='a',
                     no=1,
                     state=NodeState.UP)
    server_state_mock.get_nodes = MagicMock(return_value=[test_node])

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        result = client.get('/api/nodes')
        server_state_mock.get_nodes.assert_called_once()

        data = json.loads(result.data.decode("utf-8"))
        assert len(data['nodes']) == 1

        # Ensure that the state is converted to an integer
        assert isinstance(data['nodes'][0]['state'], int)
        assert data['nodes'][0]['state'] == NodeState.UP


def test_get_pods(client):
    server_state_mock = MagicMock()

    test_pod = Pod('a',
                   namespace='a',
                   num=1,
                   uid='a',
                   host_ip='0.0.0.0',
                   ip='0.0.0.0',
                   container_ids=['a'],
                   state=0,
                   labels=['a'])
    server_state_mock.get_pods = MagicMock(return_value=[test_pod])

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        # General case
        result = client.get('/api/pods')
        server_state_mock.get_pods.assert_called_once()

        data = json.loads(result.data.decode("utf-8"))
        assert len(data['pods']) == 1
        assert data['pods'][0]['num'] == 1

        # Case where namespace is empty
        result = client.get('/api/pods?namespace=')
        assert len(json.loads(result.data.decode("utf-8"))['pods']) == 1

        # Case where namespace matches the pod's namespace
        result = client.get('/api/pods?namespace=a')
        assert len(json.loads(result.data.decode("utf-8"))['pods']) == 1

        # Case where namespace is not the same as the pod's namespace
        result = client.get('/api/pods?namespace=b')
        assert len(json.loads(result.data.decode("utf-8"))['pods']) == 0


def test_update_nodes(client):
    server_state_mock = MagicMock()
    server_state_mock.start_node = MagicMock(return_value=True)
    server_state_mock.stop_node = MagicMock(return_value=True)

    test_node = Node(1, ip='0.0.0.0')
    server_state_mock.get_nodes = MagicMock(return_value=[test_node])

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        result = client.post('/api/nodes', data=json.dumps({
            'action': 'start',
            'ip': '0.0.0.0'
        }), content_type='application/json')
        assert result.status_code == 200
        assert server_state_mock.start_node.call_count == 1
        assert server_state_mock.stop_node.call_count == 0

        result = client.post('/api/nodes', data=json.dumps({
            'action': 'stop',
            'ip': '0.0.0.0'
        }), content_type='application/json')
        assert result.status_code == 200
        assert server_state_mock.start_node.call_count == 1
        assert server_state_mock.stop_node.call_count == 1


def test_update_pods(client):
    server_state_mock = MagicMock()

    test_pod = Pod('a', 'a', uid='1-1-1-1')
    server_state_mock.get_pods = MagicMock(return_value=[test_pod])
    server_state_mock.kill_pod = MagicMock(return_value=True)

    with mock.patch("powerfulseal.web.server.server_state", server_state_mock):
        result = client.post('/api/pods', data=json.dumps({
            'isForced': True,
            'uid': '1-1-1-1'
        }), content_type='application/json')
        assert result.status_code == 200
        server_state_mock.kill_pod.assert_called_once_with(test_pod, True)

        result = client.post('/api/pods', data=json.dumps({
            'uid': '1-1-1-1'
        }), content_type='application/json')
        assert result.status_code == 200
        server_state_mock.kill_pod.assert_called_with(test_pod, False)

        result = client.post('/api/pods', data=json.dumps({
            'isForced': False,
            'uid': '1-1-1-1'
        }), content_type='application/json')
        assert result.status_code == 200
        server_state_mock.kill_pod.assert_called_with(test_pod, False)



import pytest
from mock import patch, MagicMock
from powerfulseal.clouddrivers import azure_driver
from powerfulseal.node import Node, NodeState

PRIVATE_IPS = ['198.168.1.1', '198.168.1.2', '198.168.1.3']
PUBLIC_IPS = ['31.31.31.31', '41.41.41.41']
INVALID_IP = '198.168.3.1'

class VMPublicIP():
    def __init__(self,rg,name,ip):
        self.id='/subscriptions/a32253aa-111111/resourceGroups/'+rg+'/providers/Microsoft.Network/publicIPAddresses/'+name
        self.rg=rg
        self.name=name
        self.ip_address=ip

class VMPublicIPs():
    def __init__(self,rg,name,ip):
        self.theIP = VMPublicIP(rg,"bogusIP2","0.0.0.0")
        if name is not None:
            self.theIP=VMPublicIP(rg,name,ip)

    def get(self,rg,ipname):
        if rg == self.theIP.rg and ipname == self.theIP.name:
            return self.theIP
        return VMPublicIP(rg,"bogusIP2","0.0.0.0")

class VMIP():
    def __init__(self,priv_ip, rg, pub_ip_name, pub_ip):
        self.private_ip_address=priv_ip
        self.public_ip_address=None
        if pub_ip_name is not None:
            self.public_ip_address=VMPublicIP(rg, pub_ip_name,pub_ip)

class VMIPConfig():
    def __init__(self,rg,ifname,priv_ip,pub_ip_name, pub_ip):
        self.ip_configurations=[VMIP(priv_ip, rg, pub_ip_name, pub_ip)]
        self.rg = rg
        self.name = ifname

class VMNetworkInterface():
    def __init__(self, rg, ifname, priv_ip, pub_ip_name, pub_ip):
        self.theif= VMIPConfig(rg, ifname, priv_ip, pub_ip_name, pub_ip)

    def get(self,rg, name):
        if rg== self.theif.rg and name == self.theif.name:
            return self.theif
        return  VMIPConfig("foo", "bar", "10.10.10.10`", "bogusIp", "20.20.20.20")

class VMNetworkClient():
    def __init__(self, rg, ifname, priv_ip, pub_ip_name, pub_ip):
        self.network_interfaces=VMNetworkInterface(rg, ifname, priv_ip, pub_ip_name, pub_ip)
        self.public_ip_addresses=VMPublicIPs(rg, pub_ip_name, pub_ip)

class VMNetID():
    def __init__(self,rg, ifname):
        self.id='/subscriptions/a32253aa-111111/resourceGroups/'+rg+'/providers/Microsoft.Network/networkInterfaces/'+ifname 

class VMNetworkProfile():
    def __init__(self, rg, ifname):
        self.network_interfaces=[VMNetID(rg, ifname)]

class VMServerStatus():
    def __init__(self,state):
        self.code = state

class  VMServerMachine():
    def __init__(self, id, state):
        self.id = id
        self.statuses = [ VMServerStatus('ProvisioningState/succeeded'),  VMServerStatus(state) ]

    def instance_view(self,resource_group_name,vm_name):
        return self

class VMCompute():
    def __init__(self,id, state):
        self.virtual_machines = VMServerMachine(id, state)

class VMInstance():
    def __init__(self, id, name, zone, state, rg, ifname, priv_ip, pub_ip_name, pub_ip):
        self.id = id
        self.network_profile=VMNetworkProfile(rg, ifname)
        self.location = zone
        self.name = name
        self.state = state
        self.network_stuff = VMNetworkClient(rg, ifname, priv_ip, pub_ip_name, pub_ip)
        self.compute_stuff = VMCompute(id, state)

   
@pytest.fixture
def vm_instances():
    return [
        VMInstance(id="/subscriptions/a32253aa-111111/resourceGroups/MC_test-rg-2_westus/providers/Microsoft.Compute/virtualMachines/aks-workers-67219403-0", name="aks-workers-67219403-0", zone="westus", state="PowerState/running", rg="MC_test-rg-2_westus", ifname="two",priv_ip="198.168.1.1", pub_ip_name="publicIp01",pub_ip="31.31.31.31"),
        VMInstance(id="/subscriptions/a32253aa-111111/resourceGroups/MC_test-rg-2_westus/providers/Microsoft.Compute/virtualMachines/aks-workers-67219403-1", name="aks-workers-67219403-1", zone="westus", state="PowerState/stopped", rg="MC_test-rg-2_westus", ifname="four",priv_ip="198.168.1.2", pub_ip_name="publicIp02", pub_ip="41.41.41.41"),
        VMInstance(id="/subscriptions/a32253aa-111111/resourceGroups/MC_test-rg-2_westus/providers/Microsoft.Compute/virtualMachines/aks-workers-67219403-2", name="aks-workers-67219403-2", zone="westus", state="UNKNOWN", rg="MC_test-rg-2_westus", ifname="six",priv_ip="198.168.1.3", pub_ip_name=None, pub_ip=None),
]


@patch('powerfulseal.clouddrivers.azure_driver.create_connection_from_config', return_value=(None, None, None))
def test_get_by_ip_private_ip_nodes(create_connection_from_config, vm_instances):
    driver = azure_driver.AzureDriver()
    driver.remote_servers = vm_instances
    node_index = 0
    driver.network_client = vm_instances[node_index].network_stuff
    driver.compute_client =  vm_instances[node_index].compute_stuff
    nodes = driver.get_by_ip(PRIVATE_IPS[node_index])
    assert vm_instances[node_index].id is nodes.id
    assert vm_instances[node_index].location is nodes.az
    assert vm_instances[node_index].network_stuff.network_interfaces.theif.ip_configurations[0].private_ip_address == nodes.ip

@patch('powerfulseal.clouddrivers.azure_driver.create_connection_from_config', return_value=(None, None, None))
def test_get_by_ip_public_ip_nodes(create_connection_from_config, vm_instances):
    driver = azure_driver.AzureDriver()
    driver.remote_servers = vm_instances
    node_index = 1
    driver.network_client = vm_instances[node_index].network_stuff
    nodes = driver.get_by_ip(PUBLIC_IPS[node_index])
    assert vm_instances[node_index].id is nodes.id
    assert vm_instances[node_index].location is nodes.az
    assert vm_instances[node_index].network_stuff.network_interfaces.theif.ip_configurations[0].public_ip_address.ip_address== nodes.extIp

@patch('powerfulseal.clouddrivers.azure_driver.create_connection_from_config', return_value=(None, None, None))
def test_get_by_ip_no_nodes(create_connection_from_config, vm_instances):
    driver = azure_driver.AzureDriver()
    driver.remote_servers = vm_instances
    driver.network_client = vm_instances[2].network_stuff
    nodes = driver.get_by_ip(INVALID_IP)
    assert nodes is None



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading
import time

from mock import MagicMock

from powerfulseal.policy import PolicyRunner
from powerfulseal.web.server import ThreadedPolicyRunner


def test_threaded_policy_runner():
    # Mock all the methods which will be called by the policy runner
    test_node_scenario = MagicMock()
    test_node_scenario.execute = MagicMock()
    test_pod_scenario = MagicMock()
    test_pod_scenario.execute = MagicMock()
    test_inventory = MagicMock()
    test_inventory.sync = MagicMock(return_value=None)

    policy = {
        'config': {
            'minSecondsBetweenRuns': 0,
            'maxSecondsBetweenRuns': 2
        }
    }
    assert PolicyRunner.is_policy_valid(policy)

    test_node_scenario.execute.assert_not_called()
    test_pod_scenario.execute.assert_not_called()
    test_inventory.sync.assert_not_called()

    # As the threaded policy runner runs on a separate thread, this can be tested
    # by waiting a reasonably short amount of time for the runner loop to run
    # and checking whether the expected functions have been called by the runner
    policy_runner = ThreadedPolicyRunner(policy, test_inventory, None, None, None, threading.Event())
    policy_runner.node_scenarios = [test_node_scenario]
    policy_runner.pod_scenarios = [test_pod_scenario]

    policy_runner.start()
    time.sleep(3)
    policy_runner.stop()
    policy_runner.join()

    assert policy_runner.stop_event
    test_node_scenario.execute.assert_called()
    test_pod_scenario.execute.assert_called()
    test_inventory.sync.assert_called()
    assert test_node_scenario.execute.call_count == test_pod_scenario.execute.call_count == test_inventory.sync.call_count



import os
import sys
import logging
from . import AbstractDriver
from ..node import Node, NodeState

from azure.common.client_factory import get_client_from_auth_file
from azure.common.credentials import ServicePrincipalCredentials
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.compute import ComputeManagementClient

def create_connection_from_config():
    """ Creates a new Azure api connection """
    resource_client = None
    compute_client = None
    network_client = None
    try:
        cred_file_loc=os.environ['AZURE_AUTH_LOCATION']
    except KeyError:
        try:
            subscription_id = os.environ['AZURE_SUBSCRIPTION_ID']
            credentials = ServicePrincipalCredentials(
                client_id=os.environ['AZURE_CLIENT_ID'],
                secret=os.environ['AZURE_CLIENT_SECRET'],
                tenant=os.environ['AZURE_TENANT_ID']
            )
        except KeyError:
            sys.exit("No Azure Connection Defined")
        else:
           resource_client = ResourceManagementClient(credentials, subscription_id)
           compute_client = ComputeManagementClient(credentials, subscription_id)
           network_client = NetworkManagementClient(credentials, subscription_id)
    else:
        resource_client = get_client_from_auth_file(ResourceManagementClient)
        compute_client = get_client_from_auth_file(ComputeManagementClient)
        network_client = get_client_from_auth_file(NetworkManagementClient)

    return resource_client, compute_client, network_client

def get_all_ips(instance, network_client):
    """ Returns the private and public ip addresses of an Azure instances
    """
    output = []
    for interface in instance.network_profile.network_interfaces:
        if_name=" ".join(interface.id.split('/')[-1:])
        rg="".join(interface.id.split('/')[4])

        try:
            thing=network_client.network_interfaces.get(rg, if_name).ip_configurations
            for x in thing:
                private_ip=x.private_ip_address
                public_ip=None
                """print('\nifdata: ',rg,if_name,x.private_ip_address,x.public_ip_address)"""
                """ Have to extract public IP from public IP class structure...if present """
                if x.public_ip_address is not None:
                    public_ip_id=x.public_ip_address.id
                    public_ip_name=" ".join(public_ip_id.split('/')[-1:])
                    try:
                        public_ip_return = network_client.public_ip_addresses.get(rg, public_ip_name)
                        public_ip=public_ip_return.ip_address
                    except:
                        """ Ignore the exception.  return no additional values """
                        pass

                temp_pair = (private_ip, public_ip)
                output.append(temp_pair)

        except Exception as e:
            """ Ignore the exception.  return no additional values """
            pass

    return output

def server_state(compute_client, server):
    ret_state=NodeState.UNKNOWN
    try:
        tmp_rg="".join(server.id.split('/')[4])
        instance_view = compute_client.virtual_machines.instance_view(resource_group_name=tmp_rg, vm_name=server.name)
        status_code_operating = ",".join([s.code for s in instance_view.statuses]) or 'Unknown'
        if status_code_operating == 'ProvisioningState/succeeded,PowerState/running':
            ret_state=NodeState.UP
        elif status_code_operating != 'Unknown':
            ret_state=NodeState.DOWN
    except:
        """ignore"""
        pass

    return ret_state

def create_node_from_server(compute_client,server,int_ip, ext_ip):
    return Node(
        id=server.id,
        ip=int_ip,
        extIp=ext_ip or int_ip,
        az=server.location,
        name=server.name,
        state=server_state(compute_client,server)
    )

class AzureDriver(AbstractDriver):
    """
        Concrete implementation of the Azure cloud driver.
    """

    def __init__(self, cluster_rg_name=None, cluster_node_rg_name=None, cloud=None, conn=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.resource_client, self.compute_client, self.network_client = create_connection_from_config()
        self.remote_servers = []
        self.cluster_rg = cluster_rg_name
        self.cluster_node_rg = cluster_node_rg_name

    def getResourceGroups(self):
        """ Find nodeResourceGroup for the cluster.
            This can be determined by finding the resource groups that are managed_by 
            the cluater resource group ID
        """
        self.logger.info("++ Azure cluster_rg: %s", self.cluster_rg)
        self.logger.info("++ Azure cluster_node_rg: %s", self.cluster_node_rg)

        if self.cluster_node_rg is None:
            if self.cluster_rg is not None:
                """ get the nodeResouceGroup for the cluster ResourceGroup """
                crg = self.resource_client.resource_groups.get(self.cluster_rg)
                """ get resource groups managedBy the returned id e.g. crg.id == <>.managed_by """
                """ NOTE: azure python API resource selection filter does not appear to select correctly,
                So, query all resources and process here instead.
                """
                rgl = self.resource_client.resource_groups.list(None)
                for trg in rgl:
                    if trg.managed_by is not None:
                        if trg.managed_by.startswith(crg.id):
                            self.cluster_node_rg = trg.name
                            self.logger.info("++ Azure cluster_node_rg set to: %s", self.cluster_node_rg)
                            break
            else:
                """ Don't have the cluster resource group.
                    NOTE: it should be possible to determine it from the kubernetes cluster name.
                    however, the current (4/17/2019), python azure API doesn't support
                    querying aks by kubernetes cluster name.
                """
                self.logger.warning("Azure Cluster Resource Group is not specified.  Please specify as an argument.")


    def sync(self):
        """ Downloads a fresh set of nodes form the API.
        """
        self.logger.info("Synchronizing remote nodes")
        """only get the resource group for the current cluster 
        """
        self.getResourceGroups()
        self.remote_servers = []
        if self.cluster_node_rg is not None:
            self.remote_servers = list(self.compute_client.virtual_machines.list(self.cluster_node_rg))
        else:
            self.logger.warning("No Azure cluster node resource group was found, the node list may be incorrect.")
            self.remote_servers = list(self.compute_client.virtual_machines.list_all())

        self.logger.info("Fetched %s remote servers" % len(self.remote_servers))

    def get_by_ip(self, ip):
        """ Retrieve an instance of Node by its IP.
        """
        for server in self.remote_servers:
            addresses = get_all_ips(server, self.network_client)
            if not addresses:
                self.logger.warning("No ip addresses found: %s", server.__dict__)
            else:
                """ returns first node """
                for addr in addresses:
                    if addr[0] == ip or addr[1] == ip:
                        new_node=create_node_from_server(compute_client=self.compute_client, server=server, int_ip=addr[0], ext_ip=addr[1])
                        return new_node
        return None

    def stop(self, node):
        """ Stop a Node.
        """
        async_vm_start = self.compute_client.virtual_machines.power_off(self.cluster_node_rg, node.name)
        async_vm_start.wait()

    def start(self, node):
        """ Start a Node.
        """
        async_vm_restart = self.compute_client.virtual_machines.start(self.cluster_node_rg, node.name)
        async_vm_restart.wait()

    def delete(self, node):
        """ Delete a Node permanently.
        """
        async_vm_delete = self.compute_client.virtual_machines.delete(self.cluster_node_rg, node.name)
        async_vm_delete.wait()




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import datetime
from mock import MagicMock

import pytest

# noinspection PyUnresolvedReferences
from tests.fixtures import noop_scenario
from tests.fixtures import make_dummy_object, dummy_object


def test_empty_property_criteria_dont_match(noop_scenario):
    assert noop_scenario.match_property(None, None) is False


@pytest.mark.parametrize("query,should_match",[
    ("abc", True),
    ("def", True),
    ("lol", False),
])
def test_matches_list_types(noop_scenario, dummy_object, query, should_match):
    dummy_object.list_attr = ["abc", "def"]
    criterion = {
        "name": "list_attr",
        "value": query,
    }
    assert should_match == noop_scenario.match_property(dummy_object, criterion)


def test_filter_property(noop_scenario):
    ATTR_NAME = "test"
    dummy1 = make_dummy_object()
    setattr(dummy1, ATTR_NAME, "yes")
    dummy2 = make_dummy_object()
    setattr(dummy2, ATTR_NAME, "no")
    criterion = {
        "name": ATTR_NAME,
        "value": getattr(dummy2, ATTR_NAME)
    }
    assert noop_scenario.filter_property([dummy1, dummy2], criterion) == [dummy2]


@pytest.fixture
def some_datetime():
    dt = datetime.datetime.utcnow()
    return dt.replace(
        year = 2017,
        month = 12,
        day = 1,
        hour = 13,
        minute = 0,
        second = 0,
    )

@pytest.mark.parametrize("day,should_pass", [
    (1, True), # Friday
    (2, False), # Saturday
])
def test_filter_day_time_day_of_the_week(noop_scenario, some_datetime, day, should_pass):
    some_datetime = some_datetime.replace(day=day)
    print("new datetime", some_datetime)
    criterion = {
        "onlyDays": [
            "friday",
        ],
        "startTime": {
            "hour": 0,
            "minute": 0,
            "second": 0
        },
        "endTime": {
            "hour": 23,
            "minute": 59,
            "second": 59
        },
    }
    candidates = [make_dummy_object(), make_dummy_object()]
    if should_pass:
        assert noop_scenario.filter_day_time(candidates, criterion, now=some_datetime) == candidates
    else:
        assert noop_scenario.filter_day_time(candidates, criterion, now=some_datetime) == []


@pytest.mark.parametrize("hour,minute,second, should_pass", [
    (1,0,0, False), # 1am is bad
    (9,0,1, True), # just in time
    (13,0,0, True), # 1pm is good
    (17,59,14, True), # last minute chaos !
    (17,59,30, False), # just too late
    (23,0,0, False), # 11pm is bad
])
def test_filter_day_time_of_day(noop_scenario, some_datetime, hour, minute,second, should_pass):
    some_datetime = some_datetime.replace(
        hour=hour,
        minute=minute,
        second=second,
    )
    print("new datetime", some_datetime)
    criterion = {
        "onlyDays": [
            "friday",
        ],
        "startTime": {
            "hour": 9,
            "minute": 0,
            "second": 0
        },
        "endTime": {
            "hour": 17,
            "minute": 59,
            "second": 15
        },
    }
    candidates = [make_dummy_object(), make_dummy_object()]
    if should_pass:
        assert noop_scenario.filter_day_time(candidates, criterion, now=some_datetime) == candidates
    else:
        assert noop_scenario.filter_day_time(candidates, criterion, now=some_datetime) == []


def test_random_sample_size(noop_scenario):
    candidates = [make_dummy_object() for x in range(100)]
    for x in range(len(candidates)+1):
        sample = noop_scenario.filter_random_sample(candidates,{"size":x})
        assert len(sample) == x
        for elem in sample:
            assert elem in candidates


def test_random_sample_percentage(noop_scenario):
    candidates = [make_dummy_object() for x in range(100)]
    for x in range(101):
        percentage = x / 100.00
        sample = noop_scenario.filter_random_sample(candidates,{"ratio":percentage})
        assert len(sample) == int(percentage * len(candidates))
        for elem in sample:
            assert elem in candidates


def test_random_doesnt_pass_on_empty_criterion(noop_scenario):
    candidates = [make_dummy_object() for x in range(100)]
    for x in range(101):
        sample = noop_scenario.filter_random_sample(candidates,None)
        assert len(sample) == 0


@pytest.mark.parametrize("proba", [
    0.3,
    0.1,
    0.5,
    0.9
])
def test_filter_probability(noop_scenario, proba):
    random.seed(7) # make the tests deterministic
    SAMPLES = 100000
    candidates = [make_dummy_object(), make_dummy_object()]
    agg_len = 0.0
    criterion = {"probabilityPassAll": proba}
    for _ in range(SAMPLES):
        agg_len += len(noop_scenario.filter_probability(candidates, criterion))
    assert (agg_len/len(candidates))/SAMPLES == pytest.approx(proba, 0.01)


@pytest.mark.parametrize("sleep_time", [
    0, 1.0, 5
])
def test_action_wait(monkeypatch, noop_scenario, sleep_time):
    sleep_mock = MagicMock()
    monkeypatch.setattr("time.sleep", sleep_mock)
    criterion = {
        "seconds": sleep_time
    }
    noop_scenario.action_wait({}, criterion)
    assert sleep_mock.call_count == 1
    assert sleep_mock.call_args[0] == (sleep_time,)


def test_filter_mapping_uses_the_mapping(noop_scenario):
    dummy = make_dummy_object()
    items = [dummy]
    mapping = {
        "filterA": MagicMock(return_value=items),
        "filterB": MagicMock(return_value=[]),
    }
    filters = [
        {"filterA": {},},
        {"filterB": {},},
    ]
    res = noop_scenario.filter_mapping(items, filters, mapping)
    assert res == []
    assert mapping.get("filterA").call_count == 1
    assert mapping.get("filterA").call_args[0] == (items,{})
    assert mapping.get("filterB").call_count == 1
    assert mapping.get("filterB").call_args[0] == (items,{})


def test_acts_mapping_only_waits_once(noop_scenario):
    items = [make_dummy_object(), make_dummy_object()]
    mapping = {
        "wait": MagicMock(return_value=[]),
    }
    actions = [
        {"wait": {},},
    ]
    noop_scenario.act_mapping(items, actions, mapping)
    assert mapping.get("wait").call_count == 1




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import pytest
from mock import MagicMock


# noinspection PyUnresolvedReferences
from tests.fixtures import pod_scenario
from tests.fixtures import make_dummy_object


def test_matching_namespace(pod_scenario):
    a, b = make_dummy_object(), make_dummy_object()
    pod_scenario.schema = {
        "match": [
            {
                "namespace": {
                    "name": "something",
                }
            },
        ]
    }
    pod_scenario.k8s_inventory.find_pods = MagicMock(return_value=[a, b])
    matched = pod_scenario.match()
    assert matched == list(set([a, b]))
    assert pod_scenario.k8s_inventory.find_pods.call_args[1] == {"namespace": "something"}

def test_matching_deployment(pod_scenario):
    a, b = make_dummy_object(), make_dummy_object()
    pod_scenario.schema = {
        "match": [
            {
                "deployment": {
                    "name": "lol",
                    "namespace": "something",
                }
            },
        ]
    }
    pod_scenario.k8s_inventory.find_pods = MagicMock(return_value=[a, b])
    matched = pod_scenario.match()
    assert matched == list(set([a, b]))
    assert pod_scenario.k8s_inventory.find_pods.call_args[1] == {
        "deployment_name": "lol",
        "namespace": "something"
    }

def test_matching_labels(pod_scenario):
    a, b = make_dummy_object(), make_dummy_object()
    pod_scenario.schema = {
        "match": [
            {
                "labels": {
                    "selector": "yes=true",
                    "namespace": "something",
                }
            },
        ]
    }
    pod_scenario.k8s_inventory.find_pods = MagicMock(return_value=[a, b])
    matched = pod_scenario.match()
    assert matched == list(set([a, b]))
    assert pod_scenario.k8s_inventory.find_pods.call_args[1] == {
        "selector": "yes=true",
        "namespace": "something"
    }


@pytest.mark.parametrize("should_force", [
    (True,),
    (False,)
])
def test_kills_correctly_with_force(pod_scenario, should_force):
    pod_scenario.schema = {
        "actions": [
            {
                "kill": {
                    "force": should_force
                }
            },
        ]
    }
    mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 1
        },
    })
    pod_scenario.executor.execute = mock
    mock_item1 = MagicMock()
    mock_item1.container_ids = ["docker://container1"]
    mock_item2 = MagicMock()
    mock_item2.container_ids = ["docker://container1"]
    items = [mock_item1, mock_item2]
    pod_scenario.act(items)
    assert mock.call_count == 2
    template = "sudo docker kill -s {signal} {container_id}"
    for i, call in enumerate(mock.call_args_list):
        args, kwargs = call
        cmd = template.format(
            signal="SIGKILL" if should_force else "SIGTERM",
            container_id=items[i].container_ids[0].replace("docker://","")
        )
        assert args[0] == cmd


@pytest.mark.parametrize("prob", [
    0.3,
    0.1,
    0.5,
    0.9
])
def test_kills_with_probability(pod_scenario, prob):
    random.seed(12)
    sample = 10000

    pod_scenario.schema = {
        "actions": [
            {
                "kill": {
                    "probability": prob,
                }
            },
        ]
    }
    mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 0
        },
    })
    pod_scenario.executor.execute = mock
    mock_item1 = MagicMock()
    mock_item1.container_ids = ["docker://container1"]
    items = [mock_item1]

    for _ in range(sample):
        pod_scenario.act(items)

    actual_prob = float(mock.call_count / len(items)) / sample
    assert (actual_prob == pytest.approx(prob, 0.05))


def test_doesnt_kill_when_cant_find_node(pod_scenario):
    pod_scenario.schema = {
        "actions": [
            {
                "kill": {
                    "force": True
                }
            },
        ]
    }
    pod_scenario.inventory.get_node_by_ip = MagicMock(return_value=None)
    pod_scenario.logger = MagicMock()
    mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 1
        },
    })
    pod_scenario.executor.execute = mock
    mock_item1 = MagicMock()
    mock_item2 = MagicMock()
    items = [mock_item1, mock_item2]
    pod_scenario.act(items)
    assert mock.call_count == 0
    assert pod_scenario.logger.info.call_args[0] == ("Node not found for pod: %s", items[1])



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import random
import time

MIN_AGGRESSIVENESS = 1
MAX_AGGRESSIVENESS = 5


class DemoRunner:
    """
    Kills pods based on CPU core usage and memory usage to demonstrate how chaos
    can be created by killing pods which are busy.

    The CPU and memory information is retrieved from metrics-server. Then, pods are
    first sorted by their CPU usage and memory and killed based on filters.

    - Pods are first filtered based on what quintile they in: an aggressiveness of
    1 means that only the top 20% of sorted pods are considered; an aggressiveness of
    5 means that all nodes are considered.
    - Pods are then filtered based on a probability which corresponds to an
    aggressiveness level which is linearly scaled from 10% (1) to 80% (5).
    """

    HIGHEST_PROBABILITY = 0.8
    LOWEST_PROBABILITY = 0.1
    BASE_MINUTES = 10

    def __init__(self, inventory, k8s_inventory, driver, executor, metric_client,
                 min_seconds_between_runs=0, max_seconds_between_runs=300,
                 aggressiveness=MIN_AGGRESSIVENESS, logger=None, namespace=None,
                 metric_collector=None):
        self.inventory = inventory
        self.k8s_inventory = k8s_inventory
        self.driver = driver
        self.executor = executor
        self.min_seconds_between_runs = min_seconds_between_runs
        self.max_seconds_between_runs = max_seconds_between_runs
        self.aggressiveness = aggressiveness
        self.metric_client = metric_client
        self.logger = logger or logging.getLogger(__name__)
        self.namespace = None
        self.metric_collector = metric_collector

    def run(self):
        while True:
            # Filter
            all_pods = self.k8s_inventory.find_pods(self.namespace)
            self.logger.info("Found %d pods" % len(all_pods))
            pods = self.filter_pods(self.fill_metrics(all_pods))
            self.logger.info("Filtered to %d pods" % len(pods))

            # Execute
            for pod in pods:
                self.kill_pod(pod)

            # Sleep and sync
            sleep_time = int(random.uniform(self.min_seconds_between_runs, self.max_seconds_between_runs))
            self.logger.info("Sleeping for %s seconds", sleep_time)
            time.sleep(sleep_time)
            self.inventory.sync()

    def kill_pod(self, pod):
        # Find node
        node = self.inventory.get_node_by_ip(pod.host_ip)
        if node is None:
            self.logger.info("Node not found for pod: %s", pod)
            return

        # Format command
        container_id = random.choice(pod.container_ids)
        cmd = self.executor.get_kill_command(
            signal="SIGTERM",
            container_id=container_id.replace("docker://", ""),
        )

        # Execute command
        self.logger.info("Action execute '%s' on %r", cmd, pod)
        for value in self.executor.execute(cmd, nodes=[node]).values():
            if value["ret_code"] > 0:
                self.logger.info("Error return code: %s", value)
                self.metric_collector.add_pod_kill_failed_metric(pod)
            else:
                self.metric_collector.add_pod_killed_metric(pod)

    def sort_pods(self, pods):
        return sorted(pods, key=lambda pod: (pod.metrics['cpu'], pod.metrics['memory']), reverse=True)

    def fill_metrics(self, pods):
        """
        Adds a metrics variable to each Pod variable in a list of pods. Retrieves
        the metrics from metrics-server.
        """
        metrics = self.metric_client.get_pod_metrics()
        filled_pods = []
        for pod in pods:
            try:
                pod.metrics = metrics[pod.namespace][pod.name]
                filled_pods.append(pod)
            except (KeyError, ValueError):
                self.logger.error("Pod metrics missing for pod %s/%s - skipping" % (pod.namespace, pod.name))

        return filled_pods

    def filter_pods(self, pods):
        filters = [
            self.filter_top,
            self.filter_probability
        ]

        remaining_pods = pods
        for pod_filter in filters:
            remaining_pods = pod_filter(remaining_pods)

        return remaining_pods

    def filter_top(self, pods):
        num_pods = int(len(pods) * float(self.aggressiveness / float(MAX_AGGRESSIVENESS)))
        return pods[0:num_pods]

    def filter_probability(self, pods):
        boundary = self.LOWEST_PROBABILITY + \
                   ((self.HIGHEST_PROBABILITY - self.LOWEST_PROBABILITY) / (MAX_AGGRESSIVENESS - MIN_AGGRESSIVENESS)) * \
                   (self.aggressiveness - MIN_AGGRESSIVENESS)
        remaining_pods = []

        for pod in pods:
            p = random.random()
            if boundary >= p:
                remaining_pods.append(pod)

        return remaining_pods




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import ipaddress
from .node import Node, NodeState

class NodeInventory():

    def __init__(self, driver, restrict_to_groups=None, filters=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.driver = driver
        self.filters = filters or []
        self.local_ips = restrict_to_groups or {}
        self.groups = {}
        self.nodes_by_id = {}
        self.nodes_by_ip = {}
        self.azs = set()

    def get_all_nodes(self, sort_key="no"):
        nodes = self.nodes_by_id.values()
        return sorted(nodes, key=lambda x: getattr(x, sort_key))

    def get_node_by_ip(self, ip):
        return self.nodes_by_ip.get(ip, None)

    def find_nodes(self, query=None):
        """
            Universal node finder.
        """

        # all the queries and 'all' keyword
        if not query or query == 'all':
            for node in self.get_all_nodes():
                yield node
            return

        # support comma-separated queries
        if query and "," in query:
            for element in query.split(","):
                for node in self.find_nodes(element):
                    yield node
            return

        # match groups
        if query in self.groups.keys():
            for node in self.groups[query]:
                yield node
            return

        # match AZ
        if query in self.azs:
            for node in self.get_all_nodes():
                if node.az == query:
                    yield node
            return

        # match IP or ID or no or name
        for node in self.get_all_nodes():
            if (
                node.id == query
                or node.ip == query
                or node.extIp == query
                or str(node.no) == str(query)
                or node.name == query
            ):
                yield node
                return

        # match by state
        try:
            state = NodeState[query.upper()]
        except KeyError:
            pass
        else:
            for node in self.get_all_nodes():
                if node.state == state:
                    yield node

    def sync(self, driver=None):
        """
            Update the nodes based on the values returned from the driver
        """
        self.logger.info("Sync Nodes")
        driver = driver or self.driver
        driver.sync()
        counter = 0
        self.groups = {}
        self.nodes_by_id = {}
        self.nodes_by_ip = {}
        self.azs = set()

        for group, ips in sorted(self.local_ips.items()):
            self.groups[group] = []

            # different groups can have the same IPs,
            # so we need to match to the same nodes
            for ip in ips:
                self.logger.debug("IP processing: %s", ip)
                # see if we have a matching node in the cached list of nodes
                node = self.nodes_by_ip.get(ip)
                if node is None:
                    # node is not in cached list (or the list is empty), check the clouddriver
                    node = driver.get_by_ip(ip)
                if node is None: #pragma: no cover
                    # apart from IPs, we will also get hostnames here
                    # for those, debug, otherwise info
                    try:
                        # validate that this is an IP or not
                        ipaddress.ip_address(ip)
                        self.logger.debug("Couldn't match IP to any cloud node: %s", ip)
                    except:
                        self.logger.debug("Couldn't match to any cloud node: %s", ip)
                    continue

                self.nodes_by_id[node.id] = node
                self.nodes_by_ip[ip] = node
                self.groups[group].append(node)
                node.groups.append(group)
                self.azs.add(node.az)

                self.logger.debug("Node added: %s", node)

                # just for easier identification, give them numbers
                node.no = counter
                counter += 1

    def get_azs(self):
        return sorted(list(self.azs))

    def get_groups(self):
        return sorted(list(self.groups.keys()))



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import pkg_resources
from mock import mock, mock_open, MagicMock

from powerfulseal.execute import RemoteExecutor
from powerfulseal.node import NodeInventory, Node
from powerfulseal.policy import PolicyRunner
from powerfulseal.web.server import ServerState


def test_is_policy_valid_validates():
    valid_policy_path = pkg_resources.resource_filename("tests.policy", "example_config.yml")
    valid_policy = PolicyRunner.load_file(valid_policy_path)
    invalid_policy = {
        'config': {
            'minSecondsBetweenRuns': 'invalid'
        }
    }

    server_state = ServerState(valid_policy, None, None, None, None, None, None, None)
    assert server_state.is_policy_valid()

    server_state = ServerState(invalid_policy, None, None, None, None, None, None, None)
    assert not server_state.is_policy_valid()


def test_update_policy():
    test_path = 'test.yml'
    new_policy = {'test': 'test'}
    server_state = ServerState({}, None, None, None, None, None, None, test_path)

    m = mock_open()
    with mock.patch('powerfulseal.web.server.open', m, create=True):
        server_state.update_policy(new_policy)

    m.assert_called_once_with(test_path, 'w')
    assert server_state.policy == new_policy


def test_kill_pod():
    server_state = ServerState(None, NodeInventory(None), None, None, RemoteExecutor(), None, None, None)

    # Patch action of getting nodes to execute kill command on
    test_node = Node(1)
    get_node_by_ip_mock = MagicMock(return_value=test_node)
    server_state.inventory.get_node_by_ip = get_node_by_ip_mock

    # Patch action of choosing container
    execute_mock = MagicMock()
    server_state.executor.execute = execute_mock

    mock_pod = MagicMock()
    mock_pod.container_ids = ["docker://container1"]

    server_state.kill_pod(mock_pod, True)
    execute_mock.assert_called_once_with("sudo docker kill -s SIGKILL container1", nodes=[test_node])


def test_policy_runner_lifecycle():
    """
    Integration test which starts, checks status, and stops the policy runner
    """
    policy = {
        'config': {
            'minSecondsBetweenRuns': 0,
            'maxSecondsBetweenRuns': 2
        },
        'nodeScenarios': [
            {
                'name': 'Node Test'
            }
        ],
        'podScenarios': [
            {
                'name': 'Pod Test'
            }
        ]
    }
    test_inventory = MagicMock()
    test_inventory.sync = MagicMock(return_value=None)

    server_state = ServerState(policy, test_inventory, None, None, RemoteExecutor(), None, None, None)
    assert server_state.is_policy_runner_running() is False

    server_state.start_policy_runner()
    time.sleep(1)
    assert server_state.is_policy_runner_running() is True

    server_state.stop_policy_runner()
    time.sleep(1)
    assert server_state.is_policy_runner_running() is False




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import kubernetes.client
import kubernetes.config
from kubernetes.client.rest import ApiException


class K8sClient():
    """ Higher level Kubernetes client.
    """

    def __init__(self, kube_config=None, logger=None):
        if kube_config:
            kubernetes.config.load_kube_config(config_file=kube_config)
        else:
            kubernetes.config.load_incluster_config()
        self.client_corev1api = kubernetes.client.CoreV1Api()
        self.client_extensionsv1beta1api = kubernetes.client.ExtensionsV1beta1Api()

        self.logger = logger or logging.getLogger(__name__)
        self.logger.info("Initializing with config: %s", kube_config)

    def make_selector(self, key, value):
        """ Helper to support the name -> !value notation
        """
        if isinstance(value, str) and value.startswith("!"):
            op = "!="
            value = value[1:]
        else:
            op = "="
        return "{key}{op}{value}".format(key=key, op=op, value=value)

    def dict_to_selector(self, payload):
        """ Translates a dict into kubernetes label format
            https://kubernetes.io/docs/user-guide/labels/
        """
        if payload:
            return ",".join(self.make_selector(*item) for item in payload.items())

    def get_nodes_groups(self):
        """ Returns an inventory of nodes which form the Kubernetes cluster.
            Returns a dict of group name -> list of nodes.
        """
        nodes = self.list_nodes()
        groups = dict()
        for node in nodes:
            name = node.metadata.name
            labels = node.metadata.labels
            addresses = node.status.addresses
            ips = []
            if addresses:
                ips = [addr.address for addr in addresses]
            for label, value in labels.items():
                group = groups.get(value, [])
                for ip in ips:
                    if ip not in group:
                        group.append(ip)
                groups[value] = group
        return groups

    def list_nodes(self):
        """
            https://github.com/kubernetes-incubator/client-python/blob/master/kubernetes/docs/
            CoreV1Api.md##list_node
        """
        try:
            resp = self.client_corev1api.list_node()
            return resp.items
        except ApiException as e:
            self.logger.exception(e)
            raise

    def list_namespaces(self):
        """
            https://github.com/kubernetes-incubator/client-python/blob/master/kubernetes/docs/
            CoreV1Api.md#list_namespace
        """
        try:
            resp = self.client_corev1api.list_namespace()
            return resp.items
        except ApiException as e:
            self.logger.exception(e)
            raise

    def selector_or_labels(self, labels, selector):
        """ Helper to select labels over selector.
        """
        if selector is None:
            selector = ""
        if labels is not None:
            selector = self.dict_to_selector(labels)
        return selector

    def list_deployments(self, namespace, labels=None, selector=None):
        """
            https://github.com/kubernetes-incubator/client-python/blob/master/kubernetes/docs/
            ExtensionsV1beta1Api.md#list_namespaced_deployment
        """
        selector = self.selector_or_labels(labels, selector)
        return self.client_extensionsv1beta1api.list_namespaced_deployment(
            namespace=namespace,
            label_selector=selector,
        ).items

    def get_deployment(self, namespace, name):
        """
            https://github.com/kubernetes-incubator/client-python/blob/master/kubernetes/docs/
            ExtensionsV1beta1Api.md#read_namespaced_deployment
        """
        try:
            return self.client_extensionsv1beta1api.read_namespaced_deployment(
                namespace=namespace,
                name=name,
            )
        except ApiException as e:
            self.logger.exception(e)
            raise

    def list_pods(self, namespace, labels=None, deployment_name=None, selector=None):
        """
            https://github.com/kubernetes-incubator/client-python/blob/master/kubernetes/docs/
            CoreV1Api.md#list_namespaced_pod
            If deployment_name is provided, it will ignore selector, and use the deployment's
        """
        selector = self.selector_or_labels(labels, selector)
        if deployment_name:
            deployment = self.get_deployment(namespace, deployment_name)
            selector = self.dict_to_selector(deployment.spec.selector.match_labels)
        return self.client_corev1api.list_namespaced_pod(
            namespace=namespace,
            label_selector=selector,
        ).items

    def delete_pods(self, pods):
        """
            https://github.com/kubernetes-incubator/client-python/blob/master/kubernetes/docs/
            CoreV1Api.md#delete_namespaced_pod
            Delete all pods synchronously
        """
        for pod in pods:
            self.client_corev1api.delete_namespaced_pod(
                name=pod.name,
                namespace=pod.namespace,
                grace_period_seconds=0
            )



# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from openstack import connection, config
from . import AbstractDriver
from ..node import Node, NodeState


def create_connection_from_config(name=None):
    """ Creates a new open stack connection """
    occ = config.OpenStackConfig()
    cloud = occ.get_one_cloud(name)
    return connection.from_config(cloud_config=cloud)

def get_all_ips(server):
    """
        Digs out all the IPs from the server.addresses field
        of an open stack server.
    """
    output = []
    addresses = server.addresses
    for addr_list in addresses.values():
        for addr in addr_list:
            for name, val in addr.items():
                if name == "addr":
                    output.append(val)
    return sorted(output)

# https://developer.openstack.org/sdks/python/openstacksdk/users/resources/compute/v2/server.html#openstack.compute.v2.server.Server
MAPPING_STATES_STATUS = {
    "ACTIVE": NodeState.UP,
    "STOPPED": NodeState.DOWN,
    "SHUTOFF": NodeState.DOWN,
}
def server_status_to_state(status):
    return MAPPING_STATES_STATUS.get(status.upper(), NodeState.UNKNOWN)

def create_node_from_server(server):
    """ Translate OpenStack server representation into a Node object.
        Same IP for private (ip) and public IPs (extIp)
    """
    return Node(
        id=server.id,
        ip=get_all_ips(server)[-1],
        extIp=get_all_ips(server)[-1],
        az=server.availability_zone,
        name=server.name,
        state=server_status_to_state(server.status),
    )


class OpenStackDriver(AbstractDriver):
    """
        Concrete implementation of the OpenStack cloud driver.
    """

    def __init__(self, cloud=None, conn=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.conn = conn or create_connection_from_config(cloud)
        self.remote_servers = []

    def sync(self):
        """ Downloads a fresh set of nodes form the API.
        """
        self.logger.debug("Synchronizing remote nodes")
        self.remote_servers = list(self.conn.compute.servers())
        self.logger.info("Fetched %s remote servers" % len(self.remote_servers))

    def get_by_ip(self, ip):
        """ Retreive an instance of Node by its IP.
        """
        for server in self.remote_servers:
            addresses = get_all_ips(server)
            if not addresses:
                self.logger.warning("No addresses found: %s", server)
            else:
                for addr in addresses:
                    if addr == ip:
                        return create_node_from_server(server)
        return None

    def stop(self, node):
        """ Stop a Node.
        """
        self.conn.compute.stop_server(node.id)

    def start(self, node):
        """ Start a Node.
        """
        self.conn.compute.start_server(node.id)

    def delete(self, node):
        """ Delete a Node permanently.
        """
        self.conn.compute.delete_server(node.id)





# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import pkg_resources
from mock import MagicMock

from powerfulseal.policy import PolicyRunner


def test_default_policy_validates():
    assert PolicyRunner.is_policy_valid(PolicyRunner.DEFAULT_POLICY)


def test_example_config_validates():
    filename = pkg_resources.resource_filename("tests.policy", "example_config.yml")
    policy = PolicyRunner.load_file(filename)
    assert PolicyRunner.is_policy_valid(policy)


def test_parses_a_whole_config_correctly(monkeypatch):
    sleep_mock = MagicMock()
    monkeypatch.setattr("time.sleep", sleep_mock)
    filename = pkg_resources.resource_filename("tests.policy", "example_config2.yml")
    policy = PolicyRunner.load_file(filename)
    inventory = MagicMock()
    k8s_inventory = MagicMock()
    driver = MagicMock()
    executor = MagicMock()
    LOOPS = policy.get("config").get("loopsNumber")
    nodes, pods = PolicyRunner.run(policy, inventory, k8s_inventory, driver, executor)
    assert sleep_mock.call_count == LOOPS
    for call in sleep_mock.call_args_list:
        args, _ = call
        assert 77 <= args[0] <= 100
    assert inventory.sync.call_count == LOOPS
    assert len(nodes) == 2
    assert len(pods) == 1



# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from mock import MagicMock

# noinspection PyUnresolvedReferences
from tests.fixtures import node_scenario
from tests.fixtures import make_dummy_object


def test_matching_matches(node_scenario):
    a, b = make_dummy_object(), make_dummy_object()
    a.attr = "a - this should match"
    b.attr = "b - this won't"
    node_scenario.schema = {
        "match": [
            {
                "property": {
                    "name": "attr",
                    "value": "a.*"
                }
            },
        ]
    }
    node_scenario.inventory.find_nodes = MagicMock(return_value=[a, b])
    matched = node_scenario.match()
    assert matched == [a]


def test_matching_returns_things_once_if_multimatch(node_scenario):
    a, b = make_dummy_object(), make_dummy_object()
    a.attr = "a - this should match"
    b.attr = "b - this won't"
    node_scenario.schema = {
        "match": [
            {
                "property": {
                    "name": "attr",
                    "value": "a.*"
                }
            },
            {
                "property": {
                    "name": "attr",
                    "value": ".*"
                }
            },
        ]
    }
    node_scenario.inventory.find_nodes = MagicMock(return_value=[a, b])
    matched = node_scenario.match()
    assert len(matched) == 2
    assert a in matched
    assert b in matched


@pytest.mark.parametrize("attr", [
    "start",
    "stop"
])
def test_calls_start_on_act(node_scenario, attr):
    node_scenario.schema = {
        "actions": [
            {
                attr: {
                }
            },
        ],
    }
    mock_item = MagicMock()
    mock_item.uid = '1'
    mock_item.name = 'node1'
    items = [mock_item, mock_item]
    node_scenario.act(items)
    method = getattr(node_scenario.driver, attr)
    assert method.call_count == 2
    for i, call in enumerate(method.call_args_list):
        args, kwargs = call
        assert args[0] is items[i]


@pytest.mark.parametrize("attr", [
    "start",
    "stop"
])
def test_calls_start_on_act_raising_exception_dont_bubble(node_scenario, attr):
    node_scenario.schema = {
        "actions": [
            {
                attr: {
                }
            },
        ],
    }
    mock_item = MagicMock()
    mock_item.uid = '1'
    mock_item.name = 'node1'
    items = [mock_item, mock_item]
    method = getattr(node_scenario.driver, attr)
    method.side_effect = Exception("something bad")
    node_scenario.logger = MagicMock()
    node_scenario.act(items)
    assert method.call_count == 2
    for i, call in enumerate(method.call_args_list):
        args, kwargs = call
        assert args[0] is items[i]
    assert node_scenario.logger.exception.call_count == 2


def test_action_execute_called_correctly(node_scenario):
    node_scenario.schema = {
        "actions": [
            {
                "execute": {
                    "cmd": "echo lol",
                }
            },
        ]
    }
    mock = MagicMock(return_value={
        "some ip": {
            "ret_code": 1
        },
    })
    node_scenario.executor.execute = mock
    mock_item = MagicMock()
    mock_item.uid = '1'
    mock_item.name = 'node1'
    items = [mock_item, mock_item]
    node_scenario.act(items)
    assert mock.call_count == 2
    for i, call in enumerate(mock.call_args_list):
        args, kwargs = call
        assert args[0] == "echo lol"
        assert kwargs["nodes"] == [items[i]]



# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import time

import jsonschema
import yaml
import pkgutil
import logging
from .pod_scenario import PodScenario
from .node_scenario import NodeScenario

logger = logging.getLogger(__name__)


class PolicyRunner():
    """ Reads, validates and executes a JSON schema-compliant policy
    """
    DEFAULT_POLICY = {}

    @classmethod
    def get_schema(cls):
        """ Reads the schema from the file
        """
        data = pkgutil.get_data(__name__, "ps-schema.json")
        return yaml.safe_load(data)

    @classmethod
    def load_file(cls, filename):
        with open(filename, "r") as f:
            return yaml.safe_load(f.read())

    @classmethod
    def is_policy_valid(cls, policy, schema=None):
        schema = schema or cls.get_schema()
        try:
            jsonschema.validate(policy, schema)
        except jsonschema.ValidationError as error:
            logger.info(error)
            return False
        return True

    @classmethod
    def run(cls, policy, inventory, k8s_inventory, driver, executor,
            metric_collector=None):
        """ Runs a policy forever
        """
        config = policy.get("config", {})
        wait_min = config.get("minSecondsBetweenRuns", 0)
        wait_max = config.get("maxSecondsBetweenRuns", 300)
        loops = int(config.get("loopsNumber", -1))
        node_scenarios = [
            NodeScenario(
                name=item.get("name"),
                schema=item,
                inventory=inventory,
                driver=driver,
                executor=executor,
                metric_collector=metric_collector
            )
            for item in policy.get("nodeScenarios", [])
        ]
        pod_scenarios = [
            PodScenario(
                name=item.get("name"),
                schema=item,
                inventory=inventory,
                k8s_inventory=k8s_inventory,
                executor=executor,
                metric_collector=metric_collector
            )
            for item in policy.get("podScenarios", [])
        ]
        for _ in range(0, loops):
            for scenario in node_scenarios:
                scenario.execute()
            for scenario in pod_scenarios:
                scenario.execute()
            sleep_time = int(random.uniform(wait_min, wait_max))
            logger.info("Sleeping for %s seconds", sleep_time)
            time.sleep(sleep_time)
            inventory.sync()
        return node_scenarios, pod_scenarios
